Attribute VB_Name = "CPU65C102"
' Rockwell 65C102 emulation and CPU-related stuff.
'
' Project: FileStore [E01S emulator]
' Module : CPU65C102
'
' Version: 0.00
' Date   : 24th July 2010
' Started: 11th April 2009
'
' Author : Rick Murray, heyrick1973 -at- yahoo -dot- co -dot- uk
'          http://www.heyrick.co.uk/econet/
'
'
' Licence:
'
'   Copyright 2010 Rick Murray
'
'   Licensed under the EUPL, Version 1.1 ONLY (the "Licence");
'   You may not use this work except in compliance with the Licence.
'
'   You may obtain a copy of the Licence at:
'
'     http://ec.europa.eu/idabc/eupl5
'
'   Unless required by applicable law or agreed to in writing, software
'   distributed under the Licence is distributed on an "AS IS" basis,
'   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
'   implied. See the Licence for the specific language governing
'   permissions and limitations under the Licence.
'
'
'   Rick's note: The EUPL is "compatible" with GPL v2, but *NOT* the more
'                restrictive GPL v3. If you will be involving GPL code
'                in this project, or using some of this code in a GPL
'                project, you should specify the use of ONLY the GPL v2
'                (and not the boilerplate "GPL 2 or any later...").
'

Option Explicit


' In order to see what the processor emulation is ACTUALLY doing,
' you should set the Conditional Compilation Argument
'   TRACE=1
' (Project -> FileStore Properties -> "Make" tab)
' Then, all activity will be written to the file "trace.txt",
' but please note that the startup EPROM copy sequence ALONE will
' write about 2MiB of logging, and you'll take quite a speed hit.
'
' If you should define TRACEMORE=1 then you will have the CPU
' status added to each line, in the form:
'    ---I--Z  X=&00  Y=&00  A=&00  SP=&00
' However the startup EPROM copy will require around 5Mb, trace
' files will easily get obscenely large, and the speed penalty
' will be SEVERE.
'
'
' The CPU core, as of 2010/07/24, passes all of the verification
' test scripts. This doesn't mean it is complete, but it's a
' good start! ;-)
'


Public Cycles As Long     ' other hardware uses it for update sync
Public PrevCycles As Long ' previous cycles, for hardware polling delays

Public ExecHalt As Boolean


Private Type CPUFlags
  C As Boolean   ' Carry
  Z As Boolean   ' Zero
  I As Boolean   ' Interrupt disable
  D As Boolean   ' Decimal (BCD) mode
  B As Boolean   ' BRK command [this is 'virtual', see BRK op]
  V As Boolean   ' oVerflow
  N As Boolean   ' Negative
End Type

' Note on register values:
'   Using a 6502 simulator and single-stepping, it seems apparent
'   that the values 'wrap', so if X is 1 and you DEX twice, X
'   will then be 255. Likewise for the stack pointer.
'   Unfortunately VB using Byte data type can't be told to
'   self-wrap (it will report an "overflow" error), hence the
'   various "And 255" stuff you'll see around.

Private Type CPURegs
  X As Byte
  Y As Byte
  A As Byte        ' Accumulator
  SP As Byte       ' Stack Pointer
  PS As CPUFlags   ' Processor Status
  PC As Long       ' not Integer as int is signed...
  NMI As Boolean   ' is TRUE if an NMI has occurred
  Temp As Long     ' (internal) temporary variable for ops
  Addr As Long     ' (internal) address for loading stuff into/from
  PrevPC As Long   ' (internal) for knowing PREVIOUS location
  Abort As Boolean ' (internal) abort execution loop now...
End Type
Private CPU As CPURegs

' Addressing modes and cycle counts, as config. string and data
Private OpC(0 To 15) As String
Private CyC(0 To 15) As String
Private AddrModes(0 To 255) As Byte
Private CycCounts(0 To 255) As Byte

#If DEMON = 0 Then
' Execution history stack, for speed it's a circular list.
Private ExecStack(1 To 8) As Long
Private ExecBase As Integer
' Ditto for call/jump stack
Private JumpStack(1 To 8, 1 To 2) As Long
Private JumpBase As Integer
#End If

' Disassembly lookup table
Private DisTable As String

' The vectors are public so others can call them.
Public Const NMIVEC As Long = &HFFFA&
Public Const RSTVEC As Long = &HFFFC&
Public Const IRQVEC As Long = &HFFFE&
Public Const BRKVEC As Long = &H1234&


Private Const FCAR As Byte = 1   ' Carry
Private Const FZRO As Byte = 2   ' Zero
Private Const FIRQ As Byte = 4   ' IRQ disable
Private Const FDEC As Byte = 8   ' Decimal mode
Private Const FBRK As Byte = 16  ' BRK command
Private Const FALW As Byte = 32  ' (always set)
Private Const FOVR As Byte = 64  ' Overflow
Private Const FNEG As Byte = 128 ' Negative

Private Const AXXX As Integer = 0  ' Invalid instruction
Private Const AACC As Integer = 1  ' Accumulator
Private Const AIMM As Integer = 2  ' Immediate
Private Const AABS As Integer = 3  ' Absolute
Private Const AZPA As Integer = 4  ' Zero Page (absolute)
Private Const AZPX As Integer = 5  ' Zero Page + X
Private Const AZPY As Integer = 6  ' Zero Page + Y
Private Const AABX As Integer = 7  ' Absolute Indexed X
Private Const AABY As Integer = 8  ' Absolute Indexed Y
Private Const AIMP As Integer = 9  ' Implied
Private Const AREL As Integer = 10 ' Relative
Private Const AINX As Integer = 11 ' Indexed Indirect X
Private Const AINY As Integer = 12 ' Indirect Indexed Y
Private Const AABI As Integer = 13 ' Absolute Indirect
Private Const AIAI As Integer = 14 ' Indexed Abs. Indirect [CMOS]
Private Const AZPR As Integer = 15 ' Zero Page (relative)  [CMOS]
Private Const AIND As Integer = 16 ' Indirect              [CMOS]

Private Enum AbortReason
  INSTRUNKNOWN = 1
  ADDRUNKNOWN = 2
  INSTRUNIMPL = 3
End Enum
  
#If TRACE = 1 Then
Private TraceOp As String
Private TraceAM As String  ' AM = Addressing Mode
Private TracePC As Long
Private TraceFP As Integer
Private TraceTM As String  ' TM = TeMp
#End If



Public Sub CPU_Initialise()
  ' Set up all the CPU stuff...

  frmInit.lblStatus = "Initialising; building CPU lookup tables..."
  frmInit.lblStatus.Refresh
  
  ' This is the addressing mode table. The format is a sequence
  ' of three-character codes, which match the const declarations
  ' above, if you drop the leading A (which was necessary because
  ' "ABS" is a VB keyword).
  Dim Recurse As Integer
    
  '  0x-3x    0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
  OpC(&H0) = "IMPINXXXXXXXZPAZPAZPAZPAIMPIMMACCXXXABSABSABSZPR"
  OpC(&H1) = "RELINYINDXXXZPAZPXZPXZPAIMPABYACCXXXABSABXABXZPR"
  OpC(&H2) = "ABSINXXXXXXXZPAZPAZPAZPAIMPIMMACCXXXABSABSABSZPR"
  OpC(&H3) = "RELINYINDXXXZPXZPXZPXZPAIMPABYACCXXXABXABXABXZPR"
  '  4x-7x    0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
  OpC(&H4) = "IMPINXXXXXXXXXXZPAZPAZPAIMPIMMACCXXXABSABSABSZPR"
  OpC(&H5) = "RELINYINDXXXXXXZPXZPXZPAIMPABYIMPXXXXXXABXABXZPR"
  OpC(&H6) = "IMPINXXXXXXXZPAZPAZPAZPAIMPIMMACCXXXABIABSABSZPR"
  OpC(&H7) = "RELINYINDXXXZPAZPXZPXZPAIMPABYIMPXXXIAIABXABXZPR"
  '  8x-Bx    0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
  OpC(&H8) = "RELINXXXXXXXZPAZPAZPAZPAIMPIMMIMPXXXABSABSABSZPR"
  OpC(&H9) = "RELINYINDXXXZPXZPXZPYZPAIMPABYIMPXXXABSABXABXZPR"
  OpC(&HA) = "IMMINXIMMXXXZPAZPAZPAZPAIMPIMMIMPXXXABSABSABSZPR"
  OpC(&HB) = "RELINYINDXXXZPXZPXZPYZPAIMPABYIMPIMMABXABXABYZPR"
  '  Cx-Fx    0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
  OpC(&HC) = "IMMINXXXXXXXZPAZPAZPAZPAIMPIMMIMPXXXABSABSABSZPR"
  OpC(&HD) = "RELINYINDXXXXXXZPXZPXZPAIMPABYIMPXXXXXXABXABXZPR"
  OpC(&HE) = "IMMINXXXXXXXZPAZPAZPAZPAIMPIMMIMPXXXABSABSABSZPR"
  OpC(&HF) = "RELINYINDXXXXXXZPXZPXZPAIMPABYIMPXXXXXXABXABXZPR"

  CPU_DecodeAddrModes


  ' This is the cycle count table. The format ought to be pretty
  ' obvious. It is a raw cycle count. All the stuff with "Add 1
  ' if <blah>" is currently unimplemented; but if/when it is
  ' added, then it will be dealt with at point (because it is
  ' usually a case of "if page boundary crossed"...).

  ' Cycle counts 0x-3x
  CyC(&H0) = "7 6 0 0 5 3 5 5 3 2 2 0 6 4 6 5"
  CyC(&H1) = "2 5 5 0 5 4 6 5 2 4 2 0 6 4 7 5"
  CyC(&H2) = "6 6 0 0 3 3 5 5 4 2 2 0 4 4 6 5"
  CyC(&H3) = "2 5 5 0 4 4 6 5 2 4 2 0 4 4 7 5"
  ' Cycle counts 4x-7x
  CyC(&H4) = "6 6 0 0 0 3 5 5 3 2 2 0 3 4 6 5"
  CyC(&H5) = "2 5 5 0 0 4 6 5 2 4 3 0 0 4 7 5"
  CyC(&H6) = "6 6 0 0 3 3 5 5 4 2 2 0 6 4 6 5"
  CyC(&H7) = "2 5 5 0 4 4 6 5 2 4 4 0 6 4 7 5"
  ' Cycle counts 8x-Bx
  CyC(&H8) = "0 6 0 0 3 3 3 5 2 2 2 0 4 4 4 5"
  CyC(&H9) = "2 6 5 0 4 4 4 5 2 5 2 0 4 5 5 5"
  CyC(&HA) = "2 6 2 0 3 3 3 5 2 2 2 0 4 4 4 5"
  CyC(&HB) = "2 5 5 0 4 4 4 5 2 4 2 4 4 4 4 5"
  ' Cycle counts Cx-Fx
  CyC(&HC) = "2 6 0 0 3 3 5 5 2 2 2 0 4 4 6 5"
  CyC(&HD) = "2 5 5 0 0 4 6 5 2 4 3 0 0 4 7 5"
  CyC(&HE) = "2 6 0 0 3 3 5 5 2 2 2 0 4 4 6 5"
  CyC(&HF) = "2 5 5 0 0 4 6 5 2 4 4 0 0 4 7 5"
  
  CPU_DecodeCycCounts
  
  
  ' Set up the disassembly table. It's just a really long string...
  DisTable = "BRKORA??????TSBORAASLRMBPHPORAASL???TSBORAASLBBR" & _
             "BPLORAORA???TRBORAASLRMBCLCORAINC???TRBORAASLBBR" & _
             "JSRAND??????BITANDROLRMBPLPANDROL???BITANDROLBBR" & _
             "BMIANDAND???BITANDROLRMBSECANDDEC???BITANDROLBBR" & _
             "RTIEOR?????????EORLSRRMBPHAEORLSR???JMPEPRLSRBBR" & _
             "BVCEOREOR??????EORLSRRMBCLIEORPHY??????EORLSRBBR" & _
             "RTSADC??????STZADCRORRMBPLAADCROR???JMPADCRORBBR" & _
             "BVSADCADC???STZADCRORRMBSEIADCPLY???JMPADCRORBBR" & _
             "BRASTA??????STYSTASTXSMBDEYBITTXA???STYSTASTXBBS" & _
             "BCCSTASTA???STYSTASTXSMBTYASTATXS???STZSTASTZBBS" & _
             "LDYLDALDX???LDYLDALDXSMBTAYLDATAX???LDYLDALDXBBS" & _
             "BCSLDALDA???LDYLDALDXSMBCLVLDATSXEMULDYLDALDXBBS" & _
             "CPYCMP??????CPYCMPDECSMBINYCMPDEX???CPYCMPDECBBS" & _
             "BNECMPCMP??????CMPDECSMBCLDCMPPHX??????CMPDECBBS" & _
             "CPXSBC??????CPXSBCINCSMBINXSBCNOP???CPXSBCINCBBS" & _
             "BEQSBCSBC??????SBCINCSMBSEDSBCPLX??????SBCINCBBS"
  
  
  ' Start trace?
#If TRACE = 1 Then
  TraceFP = FreeFile
  If (Len(Dir(App.Path & "\trace.txt")) <> 0) Then
    Kill App.Path & "\trace.txt"
  End If
  Open App.Path & "\trace.txt" For Binary Access Write As TraceFP
  TraceTM = "FileStore CPU trace file - " & _
            Format(Now, "General Date") & vbCrLf & _
            String$(24, "=") & vbCrLf & vbCrLf
  Put #TraceFP, , TraceTM
#End If

  frmInit.lblStatus = "Initialising; resetting CPU..."
  frmInit.lblStatus.Refresh
  CPU_Reset
End Sub


Public Sub CPU_Execute(Loopy As Boolean)
  ' Executes instructions. At the moment it will execute 128
  ' instructions at a time, if "Loopy" is True. If False, we only
  ' execute ONE instruction.
  '
  ' The reason for 128 ops, which may really play havoc with
  ' interrupt latency (!), is because on a real machine every
  ' device is working in parallel. On an emulator, we are really
  ' only doing one thing at a time.
  ' There are two obvious approaches:
  '   1. Multiple threads. Go up a magnitude in code complexity.
  '   2. Check hardware more often. Equals slower.
  ' So we will just 'do' 128 instructions. If this is a hardship
  ' we can fine-tune the count to better balance responsiveness
  ' with speed.
  ' ##TODO## This might appear to totally f--- with the FDC that
  '          needs NMIs for very low interrupt latency (so low,
  '          in fact, that we DISABLE Econet interrupts during
  '          FDC access!) however our emulator will not be that
  '          specific in timing. It'll wait forever for the
  '          firmware to get around to reading the next byte and
  '          it will never return a "you're too damn SLOW" error.
  ' ##TODO## Check this - maybe add a hook to FDC's "IsBusy" to
  '          throttle us back to poll hardware more frequently?
  '
  '
  ' CPU emulation notes:
  '   1. We are NOT cycle-specific, but we do count cycles. Where
  '      we fail in cases where one or two cycles is added for
  '      spanning page boundaries. ##TODO##
  '
  '   2. There is CURRENTLY no support for BCD maths mode; this
  '      will only be added if the FS/MOS code require it, which
  '      I doubt. :-)
  '      ##TODO - maybe##
  '
  '   3. It is a 65C102 emulation, so will not behave the same
  '      as an NMOS 6502 (i.e. no CPU bugs!).
  '      [well, excepting those in the emulation itself! :-)]
  '
  '   4. We don't support ANY undefined instructions. The CMOS
  '      processor defines all of the undefined instructions as
  '      being NOPs, however we will fault them because they
  '      are not true NOPs, they have an instruction 'length'.
  '      ##TODO## Implement?
  '      For reference:  1 byte: None.
  '                      2 byte: &x2 / &x3 / &x4 / &xB opcodes
  '                      3 byte: &xC opcodes
  '                      If BBR/BBS unused, &xF is a 3-byter.
  '
  '  4b. We are NOT currently supporting RMBx/SMBx/BBRx/BBSx as
  '      these appear to be Rockwell-specific instructions. You
  '      will see a prompt if these are executed so I can get
  '      around to adding them...
  '      ##TODO##
  '
  '   5. When an ADDRESS is required, it will be found in the
  '      variable CPU.Addr.
  '      When a BYTE read from memory is required, it will be
  '      in the variable CPU.Temp. This is performed after we
  '      read the opcode and before anything is done about it,
  '      by way of the CPU_PreFetch routine.
  '
  '   6. It is quite obvious that CPU.Addr may contain data in
  '      other cases (usually the address of where CPU.Temp was
  '      loaded from). This is an internal artefact and should
  '      NOT be relied upon.
  '
  '   7. CPU.Addr and CPU.Temp will hold the magic value -1234
  '      if they are 'unset'. This may be seen in cases such as
  '      Implied addressing; however in such a case there will
  '      be no requirement to reference them... Basically, if
  '      you ever see a "-1234" (i.e. in trace or disassembly),
  '      it's because something has screwed up or you found a bug.
  '
  '   8. The handler for branch ops is a separate function as
  '      we have numerous opcodes that do pretty much the same
  '      thing; it saves much code duplication. But, given how the
  '      prefetch works, you'll see a fair bit of duplication
  '      anyway.
  '      ##TODO## Will calling a function add a speed hit? Would
  '               it be better if we did it in-line?
  '
  '   9. Some of the awful cruftiness you may come across are
  '      because VB is too $#!+ to be able to perform binary
  '      shifting. I mean, for God's sake, my VB (5) gives its
  '      copyright as 1987-1997; who the heck writes a mainstream
  '      computer language that can't do what any half-assed
  '      four-bit processor can do! The 808x can shift a register
  '      in 2 cycles, and a memory location in 15. The 6502 can
  '      shift the accumulator in 2 cycles and a memory location
  '      in 6. The ARM can barrel-shift as a built-in part of
  '      numerous instructions without a performance hit. BBC
  '      BASIC on a prehistoric computer can shift. Any version
  '      of the C language can shift. Hell, even JavaScript can
  '      do it. But VB? Erm... Erm... Erm...
  '      Microsoft, what WERE you thinking?!
  '
  '   A. We are *NOT* an exact emulation. A real 6502, on doing
  '      the STX instruction, would WRITE to the referenced
  '      memory location TWICE. The first write is "junk", the
  '      second being the data we're interested in.
  '
  ' This is all largely based upon my 6502 emulation originally
  ' written for AmlieEm. The primary difference is that my
  ' earlier emulator used a jumptable so each instruction was in
  ' its own routine. I don't believe VB to be that smart, so it
  ' is a great big SELECT CASE structure!
  '
  '
  ' IMPORTANT:
  '   Use CPU_Push and CPU_Pull to access the hardware stack; it
  '   keeps the stack pointer up to date.
  '   There are CPU_PushWord and CPU_PullWord for doing PC.
  '   CPU_ReadAddress is for reading 16 bit values (usu. addrs!).
  '
  '   The function CPU_PreFetch is the ONLY thing permitted to
  '   alter CPU.PC (asides from branch/jump/vector/debug-trace).
  '
  '   At point of actual instruction execution, certain data
  '   will have been prefetched (according to addressing mode)
  '   and CPU.PC will have already been updated to point to the
  '   FOLLOWING instruction. This is why we have CPU.PrevPC; this
  '   will point to the CURRENT instruction, and as such it is the
  '   preferred pointer for code tracing as it is "logically"
  '   correct.
  '
  '   The address buffers ExecStack and JumpStack are simple short
  '   CIRCULAR buffers. We don't attempt to waste any time trying
  '   to make a note of the instruction being executed, we just
  '   record the address for disassembly otherwise; we also record
  '   the target address in the case of jumps.
  
  
  ' ##TODO##
  ' When we KNOW the CPUcore is good, all of the TRACE stuff is
  ' likely to be removed. [probably soon...]
  
  Dim OpCode As Byte, OpLoop As Integer, OpEnd As Integer, SBCCarry As Boolean
  
  ' Execute 128 instructions here, saves on overheads calling
  ' the CPU_Execute function so many times...
  ' However we don't do this is Loopy is False, for else we'd
  ' have no method if single-stepping code.
  If (Loopy = True) Then OpEnd = 128 Else OpEnd = 1
  For OpLoop = 1 To OpEnd

#If TRACE = 1 Then
  TracePC = CPU.PC
  TraceOp = "???"
  TraceAM = "???"
#End If

#If DEMON = 0 Then
  ' Stack the current address
  ExecStack(ExecBase) = CPU.PC
  ExecBase = ExecBase + 1
  If (ExecBase > 8) Then ExecBase = 1
#End If

  On Error GoTo CPUCore_EpicFail

  ' Exec loop is okay
  CPU.Abort = False

  ' Remember our location as reading memory will advance PC
  CPU.PrevPC = CPU.PC
    
  ' Fetch the opcode
  OpCode = Memory_Read(CPU.PC)
  
  ' Work out addressing mode and get address/data as applicable
  CPU_PreFetch OpCode
  
  ' Now update the cycle count (raw instruction; corrections for
  ' spanning page boundary (etc) will probably need to be
  ' applied in the PreFetch).
  Cycles = Cycles + CycCounts(OpCode)
  
  ' Because of how the PreFetch works, we have now loaded up the
  ' CPU.Temp register (data) and the CPU.Addr register (address)
  ' as applicable, so the same code will now serve for the same
  ' instructions. To see what I mean, as it isn't the sort of
  ' thing I can explain easily, just look at the Select Case
  ' structure below! (it is sorted alphabetically by opcode)
  
  Select Case OpCode
    Case &H61, &H65, &H69, &H6D, &H71, &H72, &H75, &H79, &H7D:
      ' ADC        ADd with Carry
      '            A,fC = A + mem + fC
      '            Affects CZVN
#If TRACE = 1 Then
      TraceOp = "ADC"
#End If
      
      If (CPU.PS.D = False) Then
        ' Calculate to a temporary place - the Addr register (!)
        CPU.Addr = CPU.A + CPU.Temp + _
                   (IIf((CPU.PS.C = True), 1&, 0&))
        
        ' Assign the result
        CPU.A = CByte(CPU.Addr And 255&)
        
        ' Work out the flags
        CPU.PS.C = (CPU.Addr > 255&) ' over 255? set carry
        CPU.PS.Z = (CPU.A = 0&)
        CPU.PS.V = (CPU.A > 127&)    ' use A as 255+1 sets C, not V !
        CPU.PS.N = (CPU.Addr And 128&)
      End If
    
    
    
    Case &H21, &H25, &H29, &H2D, &H31, &H32, &H35, &H39, &H3D:
      ' AND        logical AND
      '            A = A & mem
      '            Affects ZN
      
#If TRACE = 1 Then
      TraceOp = "AND"
#End If
      
      CPU.A = CPU.A And CByte(CPU.Temp And 255&)
      CPU.PS.Z = (CPU.A = 0&)
      CPU.PS.N = (CPU.A And 128&)

      
      
    Case &H6, &HA, &HE, &H16, &H1E:
      ' ASL        Arithmetic Shift Left
      '            C = mem[7], mem = mem << 1 | A = mem << 1
      '            Affects CZN

#If TRACE = 1 Then
      TraceOp = "ASL"
#End If

      CPU.Temp = CPU.Temp * 2 ' or "<< 1" in a sane language :-)
      
      If (OpCode = &HA) Then
        ' Write to the accumulator
        CPU.A = CByte(CPU.Temp And 255&)
      End If
      
      CPU.PS.C = (CPU.Temp > 255&)
      CPU.Temp = (CPU.Temp And 255&) ' constrain to what fits into a byte
      CPU.PS.Z = (CPU.Temp = 0)
      CPU.PS.N = (CPU.Temp And 128&)
      
      If (OpCode <> &HA) Then
        ' Not accumulator, so write to memory
        Memory_Write CByte(CPU.Temp), CPU.Addr
      End If
      
      
    
    Case &HF, &H1F, &H2F, &H3F, &H4F, &H5F, &H6F, &H7F:
      ' BBR [ROCK] Branch on Bit Reset (clear)
      '            ##TODO## Not currently implemented
#If TRACE = 1 Then
      TraceOp = "BBRx"
#End If
      MsgBox "BBRx at &" + Right$("0000" & Hex$(CPU.PrevPC), 4), _
             vbExclamation + vbOKOnly, "CPU65C102 core - " & _
             "unsupported instruction used - BBRx."
      CPU_Fiddle OpCode, CPU.PrevPC, INSTRUNIMPL

    
    
    Case &H8F, &H9F, &HAF, &HBF, &HCF, &HDF, &HEF, &HFF:
      ' BBS [ROCK] Branch on Bit Set
      '            ##TODO## Not currently implemented
#If TRACE = 1 Then
      TraceOp = "BBSx"
#End If
      MsgBox "BBSx at &" + Right$("0000" & Hex$(CPU.PrevPC), 4), _
             vbExclamation + vbOKOnly, "CPU65C102 core - " & _
             "unsupported instruction used - BBSx."
      CPU_Fiddle OpCode, CPU.PrevPC, INSTRUNIMPL

    
    
    Case &H90, &HB0, &HF0, &H30, &HD0, &H10, &H80, &H50, &H70:
      ' BCC, BCS, BEQ, BMI, BNE, BPL, BRA, BVC, BVS
      '            We have one centralised "branch" handler
      CPU_Branch OpCode
      
      
    
    Case &H24, &H2C, &H34, &H3C, &H89:
      ' BIT        test memory BITs with accumulator
      '            A & mem
      '            Affects ZVN
#If TRACE = 1 Then
      TraceOp = "BIT"
#End If

      CPU.Temp = IIf(((CPU.A And CByte(CPU.Temp And 255&)) = CPU.Temp), 0, CPU.Temp)
      ' Now work out the flags
      CPU.PS.Z = (CPU.Temp = 0&)
      CPU.PS.V = (CPU.Temp And 64&)  ' bit 6 set
      CPU.PS.N = (CPU.Temp And 128&) ' bit 7 set

    
    Case &H0:
      ' BRK        BReaK
      '
      ' The fiddling with the Brk flag is because, to quote from
      ' Brad Taylor's document:
      '   No actual "B" flag exists inside the 6502's processor
      '   status register. The B flag only exists in the status
      '   flag byte pushed to the stack. Naturally, when the
      '   flags are restored (via PLP or RTI), the B bit is
      '   discarded.
      ' In view of this, this is what we do; however it is
      ' important to note that is appears to be different
      ' behaviour to the majority of 6502 emulators I have seen,
      ' where the B flag is SET after status pushed...
#If TRACE = 1 Then
      TraceOp = "BRK"
#End If
      
      CPU_Vector BRKVEC
            
    
    Case &H18:
      ' CLC        CLear Carry flag
#If TRACE = 1 Then
      TraceOp = "CLC"
#End If
      CPU.PS.C = False
      
    
    
    Case &HD8:
      ' CLD        CLear Decimal flag
      '
      ' We do not fault this as code written for the NMOS version
      ' of the processor needed to do this as the state of the
      ' flag is not defined at time of reset; so initialisation
      ' routines will frequently do a CLD as the first or second
      ' instruction of the reset code... The E01S code does it
      ' as the third instruction of the RAM-based reset vector
      ' (after EPROM copied, but this does not occur in
      ' subsequent resets as the EPROM is mapped out), which is
      ' perhaps a little odd knowing it was going to run on a
      ' CMOS version of the 6502 - perhaps it's just a habit...
#If TRACE = 1 Then
      TraceOp = "CLD"
#End If
      CPU.PS.D = False
    
    
    
    Case &H58:
      ' CLI        CLear Interrupt disable flag
#If TRACE = 1 Then
      TraceOp = "CLI"
#End If
      CPU.PS.I = False
    
    
    
    Case &HB8:
      ' CLV        CLear oVerflow flag
#If TRACE = 1 Then
      TraceOp = "CLV"
#End If
      CPU.PS.V = False
      
    
    
    Case &HC1, &HC5, &HC9, &HCD, &HD1, &HD2, &HD5, &HD9, &HDD:
      ' CMP        CoMPare memory with accumulator
      '            Flags = A - mem; result discarded
      '            Affects CZN
#If TRACE = 1 Then
      TraceOp = "CMP"
#End If
      
      CPU.PS.C = (CPU.A >= CPU.Temp)  ' C if Acc is >= mem
      CPU.PS.Z = (CPU.A = CPU.Temp)   ' Z if Acc = mem
      
      CPU.Temp = ((CLng(CPU.A) - CPU.Temp) And 255&)
      CPU.PS.N = (CPU.Temp And 128&) ' N if result b7 set

    
    
    Case &HE0, &HE4, &HEC:
      ' CPX        ComPare memory with X register
      '
      ' Identical to CMP above, but uses X instead of Acc
#If TRACE = 1 Then
      TraceOp = "CPX"
#End If

      CPU.PS.C = (CPU.X >= CPU.Temp)
      CPU.PS.Z = (CPU.X = CPU.Temp)
      
      CPU.Temp = ((CLng(CPU.X) - CPU.Temp) And 255&)
      CPU.PS.N = (CPU.Temp And 128&)

    
    
    Case &HC0, &HC4, &HCC:
      ' CPY        ComPare memory with Y register
      '
      ' Identical to CMP above, but uses Y instead of Acc
#If TRACE = 1 Then
      TraceOp = "CPY"
#End If

      CPU.PS.C = (CPU.Y >= CPU.Temp)
      CPU.PS.Z = (CPU.Y = CPU.Temp)
      
      CPU.Temp = ((CLng(CPU.Y) - CPU.Temp) And 255&)
      CPU.PS.N = ((CPU.Temp And 128&) = 128&)
    
    
    
    Case &H3A, &HC6, &HCE, &HD6, &HDE:
      ' DEC        DECrement memory (by one)
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "DEC"
#End If
      
      If (OpCode <> &H3A) Then
        ' working with memory
        CPU.Temp = ((CPU.Temp - 1&) And 255&)
        
        CPU.PS.Z = (CPU.Temp = 0&)
        CPU.PS.N = ((CPU.Temp And 128&) = 128&)

        Memory_Write CByte(CPU.Temp And 255&), CPU.Addr
      
      Else
        ' working with Accumulator [CMOS]
        CPU.A = ((CPU.A - 1) And 255&)
        CPU.PS.Z = (CPU.A = 0&)
        CPU.PS.N = ((CPU.A And 128&) = 128&)
      End If
      
    
    
    Case &HCA:
      ' DEX        DEcrement X register (by one)
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "DEX"
#End If
      
      CPU.X = ((CPU.X - 1) And 255&)
      CPU.PS.Z = (CPU.X = 0&)
      CPU.PS.N = ((CPU.X And 128&) = 128&)


    
    Case &H88:
      ' DEY        DEcrement Y register (by one)
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "DEY"
#End If
      
      CPU.Y = ((CPU.Y - 1&) And 255&)
      CPU.PS.Z = (CPU.Y = 0&)
      CPU.PS.N = ((CPU.Y And 128&) = 128&)


'    Case &HBB:
'      ' EMU        EMUlation test. Compares A with value given, will
'      '            report an on-screen message if there is any
'      '            difference. In code, behaves like CMP #&xx
'      '            Flags = A - mem
'      '            Affects CZN
'#If TRACE = 1 Then
'      TraceOp = "EMU"
'#End If
'
'      If (CPU.A <> CPU.Temp) Then
'        MsgBox "EMU instruction fault: expected &" & Right$( _
'               "00" & CStr(CPU.Temp), 2) & ", A was actually &" & _
'               Right$("00" & CStr(CPU.A), 2) & "." & vbCrLf & _
'               "PC is &" & Right$("0000" & Hex$(CPU.PC), 4) & ".", _
'               vbExclamation + vbOKOnly, "Emulation trap"
'      End If
'      CPU.PS.C = (CPU.A >= CPU.Temp)  ' C if Acc is >= mem
'      CPU.PS.Z = (CPU.A = CPU.Temp)   ' Z if Acc = mem
'
'      CPU.Temp = ((CLng(CPU.A) - CPU.Temp) And 255&)
'      CPU.PS.N = ((CPU.Temp And 128&) = 128&) ' N if result b7 set

        
    Case &H41, &H45, &H49, &H4D, &H51, &H52, &H55, &H59, &H5D:
      ' EOR        Exclusive OR memory with accumulator
      '            A = A ^ mem
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "EOR"
#End If
      
      CPU.A = CPU.A Xor CByte(CPU.Temp And 255&)
      CPU.PS.Z = (CPU.A = 0&)
      CPU.PS.N = ((CPU.A And 128&) = 128&)

    
    
    Case &H1A, &HE6, &HEE, &HFE, &HF6:
      ' INC        INCrement memory (by one)
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "INC"
#End If
      
      If (OpCode <> &H1A) Then
        ' working with memory
        CPU.Temp = ((CPU.Temp + 1&) And 255&) ' constrain to byte
      
        CPU.PS.Z = (CPU.Temp = 0&)
        CPU.PS.N = ((CPU.Temp And 128&) And 128&)
      
        Memory_Write CByte(CPU.Temp), CPU.Addr
      
      Else
        ' working with Accumulator [CMOS]
        CPU.A = ((CPU.A + 1) And 255)
        CPU.PS.Z = (CPU.A = 0&)
        CPU.PS.N = ((CPU.A And 128&) = 128&)
      End If

    
    
    Case &HE8:
      ' INX        Increment X register (by one)
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "INX"
#End If

      CPU.X = ((CPU.X + 1) And 255)

      CPU.PS.Z = (CPU.X = 0)
      CPU.PS.N = ((CPU.X And 128) And 128)
      
    
    
    Case &HC8:
      ' INY        Increment Y register (by one)
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "INY"
#End If

      CPU.Y = ((CPU.Y + 1) And 255)
      
      CPU.PS.Z = (CPU.Y = 0)
      CPU.PS.N = ((CPU.Y And 128) And 128)
    
    
    
    Case &H4C, &H6C, &H7C:
      ' JMP        JuMP
#If TRACE = 1 Then
      TraceOp = "JMP"
#End If
      
      CPU.PC = CPU.Addr ' simple, no? :-)
#If DEMON = 0 Then
      CPU_PushJump CPU.PrevPC
#End If
      
    
    Case &H20:
      ' JSR        Jump to SubRoutine
#If TRACE = 1 Then
      TraceOp = "JSR"
#End If
      
      CPU_PushWord (CPU.PC - 1) ' -1 is a quirk of the 6502
      CPU.PC = CPU.Addr
#If DEMON = 0 Then
      CPU_PushJump CPU.PrevPC
#End If
          
          
    
    Case &HA1, &HA5, &HA9, &HAD, &HB1, &HB2, &HB5, &HB9, &HBD:
      ' LDA        LoaD Accumulator from memory
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "LDA"
#End If
      
      CPU.A = CByte(CPU.Temp) ' already loaded it, so just assign

      CPU.PS.Z = (CPU.A = 0&)
      CPU.PS.N = ((CPU.A And 128&) And 128&)



    Case &HA2, &HA6, &HAE, &HB6, &HBE:
      ' LDX        LoaD X register from memory
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "LDX"
#End If
      
      CPU.X = CByte(CPU.Temp) ' already loaded it, so just assign

      CPU.PS.Z = (CPU.X = 0&)
      CPU.PS.N = ((CPU.X And 128&) And 128&)
    
    
    
    Case &HA0, &HA4, &HAC, &HB4, &HBC:
      ' LDY        LoaD Y register from memory
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "LDY"
#End If
      
      CPU.Y = CByte(CPU.Temp) ' already loaded it, so just assign

      CPU.PS.Z = (CPU.Y = 0&)
      CPU.PS.N = ((CPU.Y And 128&) And 128&)
    
    
    
    Case &H46, &H4A, &H4E, &H56, &H5E:
      ' LSR        Logical Shift Right
      '            C = mem[0], mem = mem >> 1 | A = mem >> 1
      '            Affects CZN
#If TRACE = 1 Then
      TraceOp = "LSR"
#End If
      
      CPU.PS.C = (CPU.Temp And 1&)
            
      CPU.Temp = CPU.Temp \ 2& ' or >> 1 in sanity-zone
      
      CPU.PS.Z = (CPU.Temp = 0&)
      CPU.PS.N = False
             
      If (OpCode = &H4A) Then
        ' Accumulator
        CPU.A = CByte(CPU.Temp And 255&)
      Else
        ' Memory
        Memory_Write CByte(CPU.Temp And 255&), CPU.Addr
      End If

    
    
    Case &HEA:
      ' NOP        No OPeration
      ' Does nothing, and takes two cycles to do so!
#If TRACE = 1 Then
      TraceOp = "NOP"
#End If

      
    
    Case &H1, &H5, &H9, &HD, &H11, &H12, &H15, &H19, &H1D:
      ' ORA        OR memory with accumulator
      '            A = A | mem
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "ORA"
#End If
      
      CPU.A = CPU.A Or CPU.Temp
      
      CPU.PS.Z = (CPU.A = 0&)
      CPU.PS.N = ((CPU.A And 128&) = 128&)
      
    
    
    Case &H48:
      ' PHA        PusH accumulator on to the stack
#If TRACE = 1 Then
      TraceOp = "PHA"
#End If
      
      CPU_Push CPU.A


    
    Case &H8:
      ' PHP        PusH Processor status on to the stack
      '            ##TODO## Is there a bugette in the 6502
      '            where PHP sets the 'B' flag? Is this present
      '            in the CMOS version as well?
#If TRACE = 1 Then
      TraceOp = "PHP"
#End If

      CPU.PS.B = True ' PHP pushes B set!
      CPU_Push CPU_AssembleStatus
      CPU.PS.B = False
  
    
    
    Case &HDA:
      ' PHX [CMOS] PusH X on to the stack
#If TRACE = 1 Then
      TraceOp = "PHX"
#End If
      
      CPU_Push CPU.X
    
    
    
    Case &H5A:
      ' PHY [CMOS] PusH Y on to the stack
#If TRACE = 1 Then
      TraceOp = "PHY"
#End If
      
      CPU_Push CPU.Y
      
    
    
    Case &H68:
      ' PLA        PulL Accumulator off of the stack
#If TRACE = 1 Then
      TraceOp = "PLA"
#End If
      
      CPU.A = CPU_Pull
      CPU.PS.Z = (CPU.A = 0&)
      CPU.PS.N = ((CPU.A And 128&) = 128&)
          
    
    
    Case &H28:
      ' PLP        PulL Processor status off of the stack
#If TRACE = 1 Then
      TraceOp = "PLP"
#End If
      
      CPU_DisassembleStatus CPU_Pull
          
    
    
    Case &HFA:
      ' PLX [CMOS] PulL X register off of the stack
#If TRACE = 1 Then
      TraceOp = "PLX"
#End If
      
      CPU.X = CPU_Pull
      CPU.PS.Z = (CPU.X = 0&)
      CPU.PS.N = ((CPU.X And 128&) = 128&)
    
    
    
    Case &H7A:
      ' PLY [CMOS] PulL Y register off of the stack
#If TRACE = 1 Then
      TraceOp = "PLY"
#End If
      
      CPU.Y = CPU_Pull
      CPU.PS.Z = (CPU.Y = 0&)
      CPU.PS.N = ((CPU.Y And 128&) = 128&)

    
    
    Case &H7, &H17, &H27, &H37, &H47, &H57, &H67, &H77:
      ' RMB [ROCK] Reset (clear) Memory Bit
      '            ##TODO## Not currently implemented
#If TRACE = 1 Then
      TraceOp = "RMBx"
#End If
      MsgBox "RMBx at &" + Right$("0000" & Hex$(CPU.PrevPC), 4), _
             vbExclamation + vbOKOnly, "CPU65C102 core - " & _
             "unsupported instruction used - RMBx."
      CPU_Fiddle OpCode, CPU.PrevPC, INSTRUNIMPL

    
    
    Case &H26, &H2A, &H2E, &H36, &H3E:
      ' ROL        ROtate Left
      '            mem = mem << 1, mem[0] = carry, carry = mem[7]
      '            Or A = mem << 1, mem[0] = carry, carry = mem[7]
      '            Affects CZN
#If TRACE = 1 Then
      TraceOp = "ROL"
#End If
      
      CPU.Temp = CPU.Temp * 2& ' << 1
      CPU.Temp = CPU.Temp Or (IIf((CPU.PS.C = True), 1&, 0&))

      ' retrieve the lost bit
      CPU.PS.C = (((CPU.Temp \ 256&) And 1&) = 1&)
      
      CPU.PS.Z = (CPU.Temp = 0&)
      CPU.PS.N = ((CPU.Temp And 128&) = 128&)
      
      If (OpCode = &H2A) Then
        ' Accumulator
        CPU.A = CByte(CPU.Temp And 255&)
      Else
        ' Memory
        Memory_Write CByte(CPU.Temp And 255&), CPU.Addr
      End If


    
    Case &H66, &H6A, &H6E, &H76, &H7E:
      ' ROR        ROtate Right
      '            mem = mem >> 1, mem[7] = carry, carry = mem[0]
      '            Or A = mem >> 1, mem[7] = carry, carry = mem[0]
      '            Affects CZN
#If TRACE = 1 Then
      TraceOp = "ROR"
#End If
      
      ' Push in carry value
      If (CPU.PS.C = True) Then
        CPU.Temp = CPU.Temp Or 256& ' << now 16 bit
      End If
      
      ' get the new carry bit
      CPU.PS.C = ((CPU.Temp And 1&) = 1&)
      
      CPU.Temp = CPU.Temp \ 2& ' >> 1
      
      CPU.PS.Z = (CPU.Temp = 0&)
      CPU.PS.N = ((CPU.Temp And 128&) = 128&)
            
      If (OpCode = &H6A) Then
        ' Accumulator
        CPU.A = CByte(CPU.Temp And 255&)
      Else
        ' Memory
        Memory_Write CByte(CPU.Temp And 255&), CPU.Addr
      End If


    
    Case &H40:
      ' RTI        ReTurn from Interrupt
#If TRACE = 1 Then
      TraceOp = "RTI"
#End If
      
      CPU_DisassembleStatus CPU_Pull
      CPU.PC = CPU_PullWord
#If DEMON = 0 Then
      CPU_PushJump CPU.PrevPC
#End If
      
      ' If NMI was set, release it now
      CPU.NMI = False
      
      
    
    Case &H60:
      ' RTS        ReTurn from Subroutine
#If TRACE = 1 Then
      TraceOp = "RTS"
#End If
      
      CPU.PC = CPU_PullWord
      CPU.PC = CPU.PC + 1 ' work around the +1 quirk in JSR
#If DEMON = 0 Then
      CPU_PushJump CPU.PrevPC
#End If
          
    
    
    Case &HE1, &HE5, &HE9, &HED, &HF1, &HF2, &HF5, &HF9, &HFD:
      ' SBC        SuBtract memory with acc with Carry
      '            A,fC = A - mem - (!fC)
      '            Affects CZVN
#If TRACE = 1 Then
      TraceOp = "SBC"
#End If

      If (CPU.PS.D = False) Then
        ' Calculate to a temporary place
        CPU.Addr = CPU.A - CPU.Temp - _
                   (IIf((CPU.PS.C = True), 0&, 1&))
                   ' NOTE that carry set = no subtract, else -1
        If ((CPU.Addr) < 0) Then
          ' less than zero? flag a carry has occurred, then
          ' correct the result to fit into a byte
          SBCCarry = True
          CPU.Addr = 256 + CPU.Addr ' casting is 'painful' in VB
        Else
          SBCCarry = False
        End If

        ' Carry is CLEAR if a borrow occurred, love the 6502! :-)
        CPU.PS.C = SBCCarry Xor True
                
        ' Overflow?
        If ((CPU.A And &H80) <> (CPU.Addr And &H80)) Then
          ' bit 7 is different between orig/final so sign has changed...
          CPU.PS.V = True
          ' ...except when the result is less than zero (a borrow)...
          If (SBCCarry = True) Then CPU.PS.V = False
        Else
          CPU.PS.V = False
        End If

        ' Assign the result
        CPU.A = CByte(CPU.Addr And 255&)

        ' And the last two...
        CPU.PS.N = (CPU.A > 127&)
        CPU.PS.Z = (CPU.A = 0&)
      End If
          
    
    
    Case &H38:
      ' SEC        SEt the Carry flag
#If TRACE = 1 Then
      TraceOp = "SEC"
#End If
      
      CPU.PS.C = True
          
    
    
    Case &HF8:
      ' SED        Set Decimal Flag
#If TRACE = 1 Then
      TraceOp = "SED"
#End If
      
      CPU.PS.D = True
      
      ' There is currently NO BCD emulation!
      If (MsgBox("Decimal mode was enabled, however the " & _
                 "processor emulation does not implement " & _
                 "BCD maths." & vbCrLf & vbCrLf & "Would you " & _
                 "like to force a system reset?", _
                 vbQuestion + vbYesNo) = vbYes) Then
        CPU_Reset
        Exit Sub
      End If
      
      
    
    Case &H78:
      ' SEI        SEt Interrupt disable flag
#If TRACE = 1 Then
      TraceOp = "SEI"
#End If
      
      CPU.PS.I = True
      
    
    
    Case &H87, &H97, &HA7, &HB7, &HC7, &HD7, &HE7, &HF7:
      ' SMB [ROCK] Set Memory Bit
      '            ##TODO## Not currently implemented
#If TRACE = 1 Then
      TraceOp = "SMBx"
#End If
      MsgBox "SMBx at &" + Right$("0000" & Hex$(CPU.PrevPC), 4), _
             vbExclamation + vbOKOnly, "CPU65C102 core - " & _
             "unsupported instruction used - SMBx."
      CPU_Fiddle OpCode, CPU.PrevPC, INSTRUNIMPL

    
    
    Case &H81, &H85, &H8D, &H91, &H92, &H95, &H99, &H9D:
      ' STA        STore Accumulator contents to memory
#If TRACE = 1 Then
      TraceOp = "STA"
#End If
      
      Memory_Write CPU.A, CPU.Addr
          
          
    
    Case &H86, &H8E, &H96:
      ' STX        STore X register contents to memory
#If TRACE = 1 Then
      TraceOp = "STX"
#End If
      
      Memory_Write CPU.X, CPU.Addr
          
    
    
    Case &H84, &H8C, &H94:
      ' STY        STore Y register contents to memory
#If TRACE = 1 Then
      TraceOp = "ASTY"
#End If
      
      Memory_Write CPU.Y, CPU.Addr
          
          
    
    Case &H64, &H74, &H9C, &H9E:
      ' STZ [CMOS] STore Zero
      ' I suppose it's less grief than LDA #0, STA <addr>!
#If TRACE = 1 Then
      TraceOp = "STZ"
#End If
      
      Memory_Write 0, CPU.Addr
          
          
          
    Case &HAA:
      ' TAX        Give money YOU earned to some idiots who will
      '            fritter it away on expenses like porn videos...
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "TAX"
#End If
      
      CPU.X = CPU.A
      
      CPU.PS.Z = (CPU.X = 0&)
      CPU.PS.N = ((CPU.X And 128&) = 128&)
      
    
    
    Case &HA8:
      ' TAY        Transfer the Accumulator into the Y register
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "TAY"
#End If
      
      CPU.Y = CPU.A
      
      CPU.PS.Z = (CPU.Y = 0&)
      CPU.PS.N = ((CPU.Y And 128&) = 128&)
      
          
    
    Case &H14, &H1C:
      ' TRB [CMOS] Test and Reset (clear) mem Bits with accum.
      '            AND the complement of A with mem, writes back.
      '            If A AND mem equals zero, Z is set.
      '            Affects Z (duh)
#If TRACE = 1 Then
      TraceOp = "TRB"
#End If
    
      CPU.Temp = (CPU.A Xor &HFF) And CPU.Temp
      Memory_Write CPU.Temp, CPU.Addr
      CPU.PS.Z = ((CPU.Temp And CPU.Addr) = 0&) ' secondary behaviour!
        
    
    
    Case &H4, &HC:
      ' TSB [CMOS] Test and Set memory Bits with accumulator
      '            ORs A with mem, stores back result.
      '            If A and mem are zero, Z is set.
      '            Affects Z
#If TRACE = 1 Then
      TraceOp = "TSB"
#End If
      
      CPU.Temp = CPU.A Or CPU.Temp
      Memory_Write CPU.Temp, CPU.Addr
      CPU.PS.Z = (CPU.Temp = 0&)
      
    
    
    Case &HBA:
      ' TSX        Transfer Stack pointer into the X register
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "TSX"
#End If

      CPU.X = CPU.SP
      
      CPU.PS.Z = (CPU.X = 0&)
      CPU.PS.N = ((CPU.X And 128&) = 128&)
      
      
      
    Case &H8A:
      ' TXA        Money paid to the government by dyslexics [*]
      '            Affects ZN
      '
      ' * - BTW, don't whinge about me being politically
      '     incorrect. I am dyscalculatic (or whatever the word
      '     is) so I have special dispensation. :-)
      '     Yup, this emulator was written by somebody who finds
      '     adding up a shopping list to be a struggle and who
      '     simply doesn't "get" long division... at all.
      '     That might explain a few things. <grin>
#If TRACE = 1 Then
      TraceOp = "TXA"
#End If
      
      CPU.A = CPU.X
      
      CPU.PS.Z = (CPU.A = 0&)
      CPU.PS.N = ((CPU.A And 128&) = 128&)
      
    
    
    Case &H9A:
      ' TXS        Transfer X register into the Stack pointer
#If TRACE = 1 Then
      TraceOp = "TXS"
#End If
      
      CPU.SP = CPU.X
         
         
         
    Case &H98:
      ' TYA        Transfer Y register into the Accumulator
      '            Affects ZN
#If TRACE = 1 Then
      TraceOp = "TYA"
#End If
      
      CPU.A = CPU.Y
      
      CPU.PS.Z = (CPU.A = 0&)
      CPU.PS.N = ((CPU.A And 128&) = 128&)
    
    
    Case Else:
      ' Unrecognised instruction...
      CPU_Fiddle OpCode, CPU.PrevPC, INSTRUNKNOWN
      
  End Select
  ' END of instruction dispatch, phew!


#If TRACE = 1 Then
  If (Mid$(TraceOp, 8, 1) = "&") Then
    ' there's a '&' so it's a Bxx with address
    TraceTM = "&" & Right$("0000" & Hex(TracePC), 4) & "  " & _
              TraceOp & "  " & TraceAM
  Else
    ' it's a "normal" instruction
    TraceTM = "&" & Right$("0000" & Hex(TracePC), 4) & "  " & _
              Left$(TraceOp & "      ", 6) & TraceAM
  End If
  
#If TRACEMORE = 1 Then
  TraceTM = Left$(TraceTM & Space(44), 44)
  TraceOp = ""
  If (CPU.PS.B = True) Then
    TraceOp = TraceOp & "B"
  Else
    TraceOp = TraceOp & "-"
  End If
  If (CPU.PS.C = True) Then
    TraceOp = TraceOp & "C"
  Else
    TraceOp = TraceOp & "-"
  End If
  If (CPU.PS.D = True) Then
    TraceOp = TraceOp & "D"
  Else
    TraceOp = TraceOp & "-"
  End If
  If (CPU.PS.I = True) Then
    TraceOp = TraceOp & "I"
  Else
    TraceOp = TraceOp & "-"
  End If
  If (CPU.PS.N = True) Then
    TraceOp = TraceOp & "N"
  Else
    TraceOp = TraceOp & "-"
  End If
  If (CPU.PS.V = True) Then
    TraceOp = TraceOp & "V"
  Else
    TraceOp = TraceOp & "-"
  End If
  If (CPU.PS.Z = True) Then
    TraceOp = TraceOp & "Z"
  Else
    TraceOp = TraceOp & "-"
  End If
  TraceOp = TraceOp & "  X=&" & Right$("00" & Hex(CPU.X), 2) & _
            "  Y=&" & Right$("00" & Hex(CPU.Y), 2) & "  A=&" & _
            Right$("00" & Hex(CPU.A), 2) & "  SP=&" & _
            Right$("00" & Hex(CPU.SP), 2)
  TraceTM = TraceTM & TraceOp & vbCrLf
#Else
  TraceTM = TraceTM & vbCrLf
#End If

  Put #TraceFP, , TraceTM
#End If

  ' Wasn't there a TXY (X -> Y) and/or TYX (Y -> X), or am I
  ' imagining things?
  ' I've looked at the Western Design Center W65C02S datasheet
  ' and the only additions I can see over the Rockwell
  ' 65C00-series (as emulated) appear to be WAI(t) and ST(o)P...
  
  
    ' Aborting?
    If (CPU.Abort = True) Then
      ExecHalt = True
      Exit Sub
    End If
  Next
  
  Exit Sub
  
CPUCore_EpicFail:
  MsgBox "CPU core error:" & vbCrLf & vbCrLf & Err.Description
  Resume Next
End Sub


Public Function CPU_AssembleStatus() As Byte
  ' Combine the status flags as a single byte (for stacking)
  ' (public for test script)
  
  Dim Flag As Byte
  
  Flag = FALW
  If (CPU.PS.B) Then Flag = Flag Or FBRK ' does not exist quite
  If (CPU.PS.C) Then Flag = Flag Or FCAR
  If (CPU.PS.D) Then Flag = Flag Or FDEC
  If (CPU.PS.I) Then Flag = Flag Or FIRQ
  If (CPU.PS.N) Then Flag = Flag Or FNEG
  If (CPU.PS.V) Then Flag = Flag Or FOVR
  If (CPU.PS.Z) Then Flag = Flag Or FZRO
  
  CPU_AssembleStatus = Flag
End Function


Private Sub CPU_DisassembleStatus(ByVal Flag As Byte)
  ' Split the processor status byte into flags (after stacking)
  
  CPU.PS.B = False ' force it FALSE
  CPU.PS.C = ((Flag And FCAR) = FCAR)
  CPU.PS.D = ((Flag And FDEC) = FDEC)
  CPU.PS.I = ((Flag And FIRQ) = FIRQ)
  CPU.PS.N = ((Flag And FNEG) = FNEG)
  CPU.PS.V = ((Flag And FOVR) = FOVR)
  CPU.PS.Z = ((Flag And FZRO) = FZRO)
End Sub


Private Sub CPU_Branch(ByVal OpCode As Byte)
  ' Centralised branch handling
  
  Dim BranchPoint As Long
  
  BranchPoint = CPU.PC + CPU.Addr ' make the branch
  
  Select Case OpCode
    Case &H90: ' BCC  Branch on Carry Clear
#If TRACE = 1 Then
      TraceOp = "BCC    &" & Right$("0000" & Hex(BranchPoint), 4)
#End If
      If (CPU.PS.C = False) Then CPU.PC = BranchPoint
      
    Case &HB0: ' BCS  Branch on Carry Set
#If TRACE = 1 Then
      TraceOp = "BCS    &" & Right$("0000" & Hex(BranchPoint), 4)
#End If
      If (CPU.PS.C = True) Then CPU.PC = BranchPoint

    Case &HF0: ' BEQ  Branch if EQual (zero flag set)
#If TRACE = 1 Then
      TraceOp = "BEQ    &" & Right$("0000" & Hex(BranchPoint), 4)
#End If
      If (CPU.PS.Z = True) Then CPU.PC = BranchPoint

    Case &H30: ' BMI Baby  A low-cost airline
#If TRACE = 1 Then
      TraceOp = "BMI    &" & Right$("0000" & Hex(BranchPoint), 4)
#End If
      If (CPU.PS.N = True) Then CPU.PC = BranchPoint

    Case &HD0: ' BNE  Branch if Not Equal
#If TRACE = 1 Then
      TraceOp = "BNE    &" & Right$("0000" & Hex(BranchPoint), 4)
#End If
      If (CPU.PS.Z = False) Then CPU.PC = BranchPoint

    Case &H10: ' BPL  Branch if PLus (not negative)
#If TRACE = 1 Then
      TraceOp = "BPL    &" & Right$("0000" & Hex(BranchPoint), 4)
#End If
      If (CPU.PS.N = False) Then CPU.PC = BranchPoint

    Case &H80: ' BRA  Appareil worn by women to prevent jiggling
#If TRACE = 1 Then
      TraceOp = "BRA    &" & Right$("0000" & Hex(BranchPoint), 4)
#End If
      CPU.PC = BranchPoint ' it's a byte less than a JMP, no?

    Case &H50: ' BVC  Branch if oVerflow Clear
#If TRACE = 1 Then
      TraceOp = "BVC    &" & Right$("0000" & Hex(BranchPoint), 4)
#End If
      If (CPU.PS.V = False) Then CPU.PC = BranchPoint

    Case &H70: ' BVS  Branch if oVerflow Set
#If TRACE = 1 Then
      TraceOp = "BVS    &" & Right$("0000" & Hex(BranchPoint), 4)
#End If
      If (CPU.PS.V = True) Then CPU.PC = BranchPoint
  End Select
  
#If DEMON = 0 Then
  If (CPU.PC = BranchPoint) Then
    ' jump was taken, note this...
    CPU_PushJump CPU.PrevPC
  End If
#End If
End Sub



Public Sub CPU_Vector(ByVal WhichVec As Long)
  ' Read the address at the specified hardware vector and
  ' load it into Program Counter; after pushing status...

  ' No IRQs when in NMI mode.
  If ((WhichVec = IRQVEC) And (CPU.NMI = True)) Then Exit Sub

  ' No IRQs while an IRQ is in progress.
  If ((WhichVec = IRQVEC) And (CPU.PS.I = True)) Then Exit Sub

  ' If BRK opcode...
  If (WhichVec = BRKVEC) Then
    CPU.PS.B = True
    WhichVec = IRQVEC
    CPU.PC = CPU.PC + 1 ' BRK is a TWO byte instruction!
  Else
    ' not BRK, force B flag off
    CPU.PS.B = False
  End If

  ' Push current status to the stack
  CPU_PushWord CPU.PC
  CPU_Push CPU_AssembleStatus
  
  ' Set the IRQ disable flag
  CPU.PS.I = True
  
  ' Disable the BRK flag (it is only set in the stored byte!)
  CPU.PS.B = False

  ' If NMI mode, set NMI lockout
  If (WhichVec = NMIVEC) Then CPU.NMI = True

  ' Okay, follow the vector
  CPU.PC = CPU_ReadAddress(WhichVec)
#If DEMON = 0 Then
  CPU_PushJump WhichVec ' the hardware vectors are "special"
#End If


#If TRACE = 1 Then
  ' Trace stuff
  Select Case WhichVec
    Case IRQVEC: TraceTM = "       [IRQ vector called]"
    Case NMIVEC: TraceTM = "       [NMI vector called]"
    Case RSTVEC: TraceTM = "       [RST vector called]"
    Case Else: TraceTM = "      [unknown vector called " & _
                         "- &" & Right$("0000" & _
                         Hex$(WhichVec), 4) & "]"
    TraceTM = vbCrLf & TraceTM & vbCrLf & vbCrLf
    Put #TraceFP, , TraceTM
  End Select
#End If
End Sub


Private Sub CPU_Push(ByVal Value As Byte)
  ' Pushes a byte to the processor stack.
  Memory_Write Value, 256 + CPU.SP
  CPU.SP = ((CPU.SP - 1) And 255)
End Sub


Private Function CPU_Pull() As Byte
  ' Retrieves a byte from the processor stack.
  CPU.SP = ((CPU.SP + 1) And 255)
  CPU_Pull = Memory_Read(256 + CPU.SP)
End Function


Private Sub CPU_PushWord(ByVal Value As Long)
  ' Pushes a word to the processor stack.
  Memory_Write ((Value \ 256&) And 255&), (256 + CPU.SP) ' high
  CPU.SP = ((CPU.SP - 1) And 255)
  Memory_Write (Value And 255&), (256 + CPU.SP)          ' low
  CPU.SP = ((CPU.SP - 1) And 255)
End Sub


Private Function CPU_PullWord() As Long
  ' Retrieves a word from the processor stack.
  Dim Value As Long
  
  CPU.SP = ((CPU.SP + 1) And 255)
  Value = Memory_Read(256 + CPU.SP)
  CPU.SP = ((CPU.SP + 1) And 255)
  Value = Value Or (Memory_Read(256 + CLng(CPU.SP)) * 256&)
  
  CPU_PullWord = Value
End Function


Private Function CPU_ReadAddress(ByVal Where As Long) As Long
  ' Useful function to read two bytes and generate an address.
  
  Dim Addr As Long
  
  Addr = Memory_Read(Where)
  Addr = Addr + (Memory_Read(Where + 1&) * 256&)
  
  CPU_ReadAddress = Addr
End Function


Private Sub CPU_SetZN(ByVal OffWhat As Byte)
  ' Given a byte, set the Zero and Negative flags accordingly.
  CPU.PS.Z = (OffWhat = 0)
  CPU.PS.N = (((OffWhat = 128) And 128) = 128)
End Sub


Private Function CPU_SetV() As Boolean
  ' Performs the oVerflow calculation
  '
  ' ASSUMES the Carry flag has been appropriately set, and
  ' that a working value is in CPU.Addr (*NOT* CPU.Temp).
  '
  ' oVerflow is set when:
  '   There is a carry from bit six of result to bit seven,
  '   with NO external carry.
  ' or:
  '   There is an external carry, but NO carry from bit six
  '   to bit seven of the result.
  '
  ' BBC Micro - Advanced User Guide p32 and p42
  '
  
  If ((CPU.A And &H80) = (CPU.Addr And &H80)) Then
    ' No carry in result, is carry set?
    If (CPU.PS.C = True) Then
      CPU_SetV = True
      Exit Function
    End If
  
  Else
    ' Carry in result, is carry NOT set?
    If (CPU.PS.C = False) Then
      CPU_SetV = True
      Exit Function
    End If
  End If
  
  CPU_SetV = False
End Function



Private Sub CPU_DecodeAddrModes()
  ' Address mode table is a sequence of 16 strings which contain
  ' 16 three-character codes. This is parsed and split into a
  ' simpler lookup table...
  
  Dim inner As Integer, Outer As Integer, Match As String
  Dim Result As Byte
  
  For Outer = 0 To 15
    For inner = 0 To 15
      Match = UCase$(Mid$(OpC(Outer), ((inner * 3) + 1), 3))
      
      Select Case Match
        Case "XXX": Result = AXXX
        Case "ACC": Result = AACC
        Case "IMM": Result = AIMM
        Case "ABS": Result = AABS
        Case "ZPA": Result = AZPA
        Case "ZPX": Result = AZPX
        Case "ZPY": Result = AZPY
        Case "ABX": Result = AABX
        Case "ABY": Result = AABY
        Case "IMP": Result = AIMP
        Case "REL": Result = AREL
        Case "INX": Result = AINX
        Case "INY": Result = AINY
        Case "ABI": Result = AABI
        Case "IAI": Result = AIAI
        Case "ZPR": Result = AZPR
        Case "IND": Result = AIND
      End Select
      
      AddrModes((Outer * 16) + inner) = Result
    Next
  Next
End Sub
  
  
Private Sub CPU_DecodeCycCounts()
  ' The cycle count table is a sequence of 16 strings which
  ' contain 16 numbers separated by a single space. This is
  ' split into a simpler lookup table...
  
  Dim inner As Integer, Outer As Integer
  
  For Outer = 0 To 15
    For inner = 0 To 15
      CycCounts((Outer * 16) + inner) = _
        CByte(Mid$(CyC(Outer), ((inner * 2) + 1), 1))
    Next
  Next
End Sub
  
  
Public Sub CPU_Reset()
  ' Forces a CPU reset
  Dim Recurse As Integer

#If DEMON = 0 Then
  For Recurse = 1 To 8
    ExecStack(Recurse) = -1
    JumpStack(Recurse, 1) = -1
    JumpStack(Recurse, 2) = -1
  Next
  ' the base always points to where to write the next entry,
  ' and can also be used so determine where to start reading
  ' (with the 'magic value' -1 meaning "no meaningful data",
  '  though this condition will only exist for maybe a few
  '  hundred emulated cycles at the start of the execution)
  ExecBase = 1
  JumpBase = 1
#End If

  ' most likely this reset will be a hardware reset, so reset that
  ADLC_Initialise
  FDC_Initialise
  HDC_Initialise
  Memory_Initialise  ' resets EPROM/RAM access latch
  Memory_PatchROM    ' apply in-situ patches to EPROM image
  NVRAM_Initialise
  VIA_Initialise
  
  CPU.X = 0
  CPU.Y = 0
  CPU.A = 0
  CPU.NMI = False
  CPU.SP = 0       ' does it default to zero? does it matter?
  CPU.PS.B = False
  CPU.PS.C = False
  CPU.PS.D = False
  CPU.PS.I = True  ' IRQ disable set at reset...
  CPU.PS.N = False
  CPU.PS.V = False
  CPU.PS.Z = False
  CPU.PC = CPU_ReadAddress(&HFFFC&) ' directly, not via CPU_Vector
    ' IMPORTANT: On the CMOS processor, the vector IS taken, so it
    '            is actually possible to RTI out of a reset! :-)
End Sub


Private Sub CPU_PreFetch(ByVal OpCode As Byte)
  ' Here we deal with the addressing mode related stuff.
  '
  ' What we do:
  '   Determine if this is a WRITE op (or a JUMP op).
  '   Read the parameters according to addressing mode, which
  '   will set up CPU.Addr with an address and/or CPU.Temp with
  '   a value read from memory.
  '   EXCEPTION: If it's a write or jump op, we do not read from
  '              memory - because it is not necessary; and also
  '              because silently READING address &FC08 when we
  '              are intending to WRITE to it will have rather
  '              specific unwanted effects. [it's a read
  '              sensitive location that will map out the EPROM
  '              and switch to RAM only; while WRITING to that
  '              address is how we switch on the red LED at the
  '              very start - you see why reading on a write is
  '              a Bad Thing? :-) ]
  
  Dim DoRead As Boolean, EffAddr As Long

  ' Blank the internal registers;
  ' -1234 means "unset" (we can't use -1 as rel. addr can set it)
  CPU.Addr = -1234
  CPU.Temp = -1234
  
  ' Increment PC for the opcode we've already read; remember the
  ' ONLY things that can mess with PC are here, and JSR/JMP/Bxx.
  CPU.PC = CPU.PC + 1
  
  ' Should we read memory? Yes for LDX, No for STX (etc)
  DoRead = True ' default to YES
  Select Case OpCode
    Case &H64, &H6C, &H74, &H7C, &H81, &H84: DoRead = False
      '  STZ   JMP   STZ   JMP   STA   STY
      
    Case &H85, &H86, &H8C, &H8D, &H8E, &H91: DoRead = False
      '  STA   STX   STY   STA   STX   STA
      
    Case &H92, &H94, &H95, &H96, &H99, &H9C: DoRead = False
      '  STA   STY   STA   STX   STA   STZ
    
    Case &H9D, &H9E, &H7C, &O20: DoRead = False
      '  STA   STZ   JMP   JSR
  End Select
  ' We also include support for NOT reading on JMP/JSR, as it
  ' is not necessary to perform any reads in these cases.
  ' ##TODO## Bxx branch also?
    
  
  
  ' ##TODO## Assure that we are only READING when absolutely
  '          necessary; for we may end up inadvertantly reading
  '          from a read-sensitive location, or perhaps something
  '          not even IN the memory map... when in actuality
  '          such a read was not necessary.
  
  ' ##TODO## Various instruction/addressing mode combinations
  '          cause either 1 or 2 cycles to be added if a page
  '          boundary is crossed. This should be implemented.
  
  
  Select Case AddrModes(OpCode)
    Case AXXX: ' Invalid reference - we don't support random NOPs
#If TRACE = 1 Then
      TraceTM = "!!!!!  Invalid opcode &" & Hex$(OpCode) & ", " & _
             "PC = &" & Right$("0000" & Hex$(CPU.PC), 4) & vbCrLf
      Put #TraceFP, , TraceTM
#End If
      ' we do NOT fault this here, the CPU core will do it...
      Exit Sub
    
    Case AACC: ' Accumulator
      ' The value in the accumulator is our data.
      CPU.Temp = CPU.A
#If TRACE = 1 Then
      TraceAM = " A"
#End If

    Case AIMM: ' Immediate
      ' The following byte is our data.
      CPU.Temp = Memory_Read(CPU.PC)
      CPU.PC = CPU.PC + 1
#If TRACE = 1 Then
      TraceAM = " #&" & Hex$(CPU.Temp)
#End If
      
    Case AABS: ' Absolute
      ' The following two bytes determine the address of our
      ' data.
      CPU.Addr = CPU_ReadAddress(CPU.PC)
      If (DoRead = True) Then CPU.Temp = Memory_Read(CPU.Addr)
      CPU.PC = CPU.PC + 2
#If TRACE = 1 Then
      TraceAM = " &" & Right$("0000" & Hex$(CPU.Addr), 4)
      If (DoRead = True) Then
        TraceAM = TraceAM & "    ; <= &" & Hex$(CPU.Temp)
      End If
#End If
    
    Case AZPA: ' Zero Page Absolute
      ' The following byte determines the address of our data
      ' (like Absolute, only in Page Zero)
      CPU.Addr = Memory_Read(CPU.PC)
      If (DoRead = True) Then CPU.Temp = Memory_Read(CPU.Addr)
      CPU.PC = CPU.PC + 1
#If TRACE = 1 Then
      TraceAM = " &" & Right$("00" & Hex$(CPU.Addr), 2)
      If (DoRead = True) Then
        TraceAM = TraceAM & "      ; <= &" & Hex$(CPU.Temp)
      End If
#End If
    
    Case AZPX: ' Zero Page, X
      ' The following byte, plus X, give the address of our data.
      EffAddr = Memory_Read(CPU.PC)
      CPU.Addr = EffAddr + CLng(CPU.X)
      If (DoRead = True) Then CPU.Temp = Memory_Read(CPU.Addr)
      CPU.PC = CPU.PC + 1
#If TRACE = 1 Then
      TraceAM = " &" & Right$("00" & Hex$(EffAddr), 2) & ",X"
      If (DoRead = True) Then
        TraceAM = TraceAM & "    ; <= &" & Hex$(CPU.Temp) & _
                  " @ &" & Right$("0000" & Hex$(CPU.Addr), 4)
      End If
#End If

    Case AZPY: ' Zero Page, Y
      ' The following byte, plus Y, give the address of our data.
      ' This instruction is for LDX. Everything else uses ZP,X.
      EffAddr = Memory_Read(CPU.PC)
      CPU.Addr = EffAddr + CLng(CPU.Y)
      If (DoRead = True) Then CPU.Temp = Memory_Read(CPU.Addr)
      CPU.PC = CPU.PC + 1
#If TRACE = 1 Then
      TraceAM = " &" & Right$("00" & Hex$(EffAddr), 2) & ",Y"
      If (DoRead = True) Then
        TraceAM = TraceAM & "    ; <= &" & Hex$(CPU.Temp) & _
                  " @ &" & Right$("0000" & Hex$(CPU.Addr), 4)
      End If
#End If
    
    Case AABX: ' Absolute, X
      ' The following two bytes, plus X, give our data's address.
      ' It's funny that we can LDY <addr>,X but there is no
      ' corresponding STY using this addressing mode!
      EffAddr = CPU_ReadAddress(CPU.PC)
      CPU.Addr = EffAddr + CLng(CPU.X)
      If (DoRead = True) Then CPU.Temp = Memory_Read(CPU.Addr)
      CPU.PC = CPU.PC + 2
#If TRACE = 1 Then
      TraceAM = " &" & Right$("0000" & Hex$(EffAddr), 4) & ",X"
      If (DoRead = True) Then
        TraceAM = TraceAM & "  ; <= &" & Hex$(CPU.Temp) & _
                  " @ &" & Right$("0000" & Hex$(CPU.Addr), 4)
      End If
#End If
      ' This is used, for example, by the FileStore MOS when
      ' reloading default values into the NVRAM:
      '     LDX #&0E      ' skip the RTC part
      '   .loop
      '     LDX &0531, X  ' read from table at &053F (&531+&E...)
      '     PHX
      '     JSR &028A     ' routine to fake a call to OSWORD 6
      '     PLX           ' (which writes a byte to NVRAM)
      '     INX
      '     CPX #&32      ' loop until all 32 bytes written

    Case AABY: ' Absolute, Y
      ' The following two bytes, plus Y, give our data's address.
      ' It's funny that we can LDX <addr>,Y but there is no
      ' corresponding STX using this addressing mode!
      EffAddr = CPU_ReadAddress(CPU.PC)
      CPU.Addr = EffAddr + CLng(CPU.Y)
      If (DoRead = True) Then CPU.Temp = Memory_Read(CPU.Addr)
      CPU.PC = CPU.PC + 2
#If TRACE = 1 Then
      TraceAM = " &" & Right$("0000" & Hex$(EffAddr), 4) & ",Y"
      If (DoRead = True) Then
        TraceAM = TraceAM & "  ; <= &" & Hex$(CPU.Temp) & _
                  " @ &" & Right$("0000" & Hex$(CPU.Addr), 4)
      End If
#End If
      
    Case AIMP: ' Implied
      ' No addressing mode is relevant (i.e. CLD or SEI...)
#If TRACE = 1 Then
      TraceAM = ""
#End If
    
    Case AREL: ' Relative
      ' The following byte is a relative address in the range
      ' -128 to 127... no data is read as the address is for
      ' branching.
      CPU.Addr = Memory_Read(CPU.PC)
      If (CPU.Addr > 127) Then
        CPU.Addr = (-256& + CPU.Addr) ' make it 'signed'
      End If
      CPU.PC = CPU.PC + 1
#If TRACE = 1 Then
      TraceAM = " {relative " & CStr(CPU.Addr) & "}"
#End If

    Case AINX: ' Indexed Indirect, X [aka Pre-indexed indirect]
      ' The second byte of the instruction, plus X, points to a
      ' location in page zero where the real address may be
      ' found. Data is then loaded from this eventual address...
      ' This means a list of addresses is held in page zero.
      ' Note: There's no 'Y' version of this addressing mode.
      EffAddr = Memory_Read(CPU.PC)
      CPU.Addr = EffAddr + CLng(CPU.X)
      CPU.Addr = CPU_ReadAddress(CPU.Addr)
      If (DoRead = True) Then CPU.Temp = Memory_Read(CPU.Addr)
      CPU.PC = CPU.PC + 1
#If TRACE = 1 Then
      TraceAM = " (&" & Right$("00" & Hex$(EffAddr), 2) & ",X)"
      If (DoRead = True) Then
        TraceAM = TraceAM & "  ; <= &" & Hex$(CPU.Temp) & _
                  " @ &" & Right$("0000" & Hex$(CPU.Addr), 4)
      End If
#End If
      
    Case AINY: ' Indexed Indirect, Y [aka Post-indexed indirect]
      ' Ready? Okay... The second byte of the instruction points
      ' to a location in page zero where the 'real' address may
      ' be found. The Y register is then applied to this address
      ' to create the eventual address from which we load our
      ' data. [phew!]
      ' This means a pointer to a list of addresses is held in
      ' page zero, but the actual list is located elsewhere.
      ' Note: There's no 'X' version of this addressing mode.
      EffAddr = Memory_Read(CPU.PC)
      CPU.Addr = CPU_ReadAddress(EffAddr) + CLng(CPU.Y)
      If (DoRead = True) Then CPU.Temp = Memory_Read(CPU.Addr)
      CPU.PC = CPU.PC + 1
#If TRACE = 1 Then
      TraceAM = " (&" & Right$("00" & Hex$(EffAddr), 2) & "),Y"
      If (DoRead = True) Then
        TraceAM = TraceAM & "  ; <= &" & Hex$(CPU.Temp) & _
                  " @ &" & Right$("0000" & Hex$(CPU.Addr), 4)
      End If
#End If
    
    Case AABI: ' Absolute Indirect
      ' The address given points to another address (this might
      ' seem odd, but it is useful for vectors that can be
      ' updated on-the-fly).
      CPU.Addr = CPU_ReadAddress(CPU.PC)
#If TRACE = 1 Then
      TraceAM = " (&" & Right$("0000" & Hex$(CPU.Addr), 4) & ")"
#End If
      CPU.Addr = CPU_ReadAddress(CPU.Addr)
#If TRACE = 1 Then
      TraceAM = TraceAM & "  ; <= &" & Hex$(CPU.Addr)
#End If
      ' The NMOS 6502 is bugged in that if the address spans a
      ' page boundary (like an address at &07FF), then the
      ' address is not correctly incremented, so the above would
      ' result in a read from locations &07FF and &0700. The CMOS
      ' version of the processor does it correctly; reading from
      ' &07FF then &0800. [and strictly speaking this SHOULD add
      ' an extra cycle ##TODO##]
    
    Case AIAI: ' Indexed Absolute Indirect [CMOS]
      ' This new addressing mode on the CMOS processor takes the
      ' following two bytes, plus X, to given the location of a
      ' new address. This is pretty much ABX, but we are reading
      ' in an address, not data. For JMP (ABS, X).
      EffAddr = CPU_ReadAddress(CPU.PC)
      CPU.Addr = EffAddr + CLng(CPU.X)
#If TRACE = 1 Then
      TraceAM = " (&" & Right$("0000" & Hex$(EffAddr), 4) & ",X)"
#End If
      CPU.Addr = CPU_ReadAddress(CPU.Addr)
#If TRACE = 1 Then
      TraceAM = TraceAM & "  ; <= &" & Hex$(CPU.Addr)
#End If
      CPU.PC = CPU.PC + 2

    Case AZPR: ' Zero Page Relative
      ' This new CMOS addressing mode tests the zero page
      ' location specified for bit set/clear and then takes the
      ' relative branch accordingly.
      ' Tom Walker [BBC micro mailing list] suggests that the
      ' CPUs used in Acorn machines may not do this?
      ' [msgid <840936.94599.qm@web27204.mail.ukl.yahoo.com>]
      CPU.Addr = Memory_Read(CPU.PC)
#If TRACE = 1 Then
      TraceAM = " &" & Right$("00" & Hex$(CPU.Addr), 2)
#End If
      CPU.Temp = Memory_Read(CPU.Addr) ' get ZP byte into Temp
      CPU.PC = CPU.PC + 1&
      CPU.Addr = Memory_Read(CPU.PC)
      CPU.Addr = (-128& + CPU.Addr)    ' make branch 'signed'
#If TRACE = 1 Then
      TraceAM = " {rel " & CStr(CPU.Addr) & "}  ; <= &" & _
                Hex$(CPU.Temp)
#End If
      CPU.PC = CPU.PC + 1&
    
    Case AIND: ' Indirect
      ' This new CMOS addressing mode contains a pointer to a
      ' word in page zero that points to our memory byte...
      CPU.Addr = Memory_Read(CPU.PC)
      CPU.PC = CPU.PC + 1&
#If TRACE = 1 Then
      TraceAM = " (&" & Right$("00" & Hex$(CPU.Addr), 2) & ")"
#End If
      CPU.Addr = CPU_ReadAddress(CPU.Addr)
#If TRACE = 1 Then
      TraceAM = TraceAM & "  ; <= &" & Hex$(CPU.Addr)
#End If
      If (DoRead = True) Then CPU.Temp = Memory_Read(CPU.Addr)
    
    Case Else: ' error
#If TRACE = 1 Then
      TraceTM = "!!!!!  Unknown addressing mode with opcode &" & _
                Hex$(OpCode) & ", PC = &" & Right$("0000" & _
                Hex$(TracePC), 4) & vbCrLf
      Put #TraceFP, , TraceTM
#End If
      CPU_Fiddle OpCode, (CPU.PC - 1), ADDRUNKNOWN
  End Select
  
End Sub


Public Function CPU_State() As String
  ' Build up a string of the processor status for appending to
  ' the title in frmMain.
  
  Dim Reply As String
  Reply = "PC = &" & Right$("0000" & Hex(CPU.PC), 4) & _
          ", cycles = " & CStr(Cycles) & "."

#If TRACEMORE = 1 Then
  Reply = Reply & " X=&" & Hex(CPU.X) & " Y=&" & Hex(CPU.Y) & _
          " A=&" & Hex(CPU.A)
#End If

  CPU_State = Reply
End Function


Public Function CPU_ReadPC() As Long
  ' We return the PREVIOUS location of PC as the majority of
  ' uses will want to know the address of the instruction that
  ' triggered the event. By the time we get to op dispatch, PC
  ' will be pointing to the NEXT instruction...
  CPU_ReadPC = CPU.PrevPC
  
  ' Now a hack so disassembly works right from the get-go.
  If (Cycles = 0) Then CPU_ReadPC = CPU.PC
End Function


#If DEMON = 0 Then
Private Sub CPU_PushJump(ByVal Where As Long)
  ' In this array, offset #1 is the address of the instruction
  ' that caused the jump, and offset #2 is the eventual address
  ' of PC following the jump; thus:
  '   Instruction [#1] jumped us to [#2]
  
  JumpStack(JumpBase, 1) = Where
  JumpStack(JumpBase, 2) = CPU.PC
  JumpBase = JumpBase + 1
  If (JumpBase > 8) Then JumpBase = 1
End Sub
#End If


Public Sub CPU_RefreshTrace()
  ' Updates the frmCPUState window
  Dim Recurse As Integer, MyBase As Integer, MyTxt As String
  Dim OpCode As Byte, EmTime As Double
  
  If (frmCPUState.Visible = False) Then Exit Sub
  
  With frmCPUState
  ' registers, basic
  .txtReg(0).Text = Hex$(CPU.X)
  .txtReg(1).Text = Hex$(CPU.Y)
  .txtReg(2).Text = Hex$(CPU.A)
  .txtReg(3).Text = Hex$(CPU.SP)
  
  ' registers, complex
  .lblRegDec(0).Caption = CPU_RegDec(CPU.X)
  .lblRegDec(1).Caption = CPU_RegDec(CPU.Y)
  .lblRegDec(2).Caption = CPU_RegDec(CPU.A)
  .lblRegDec(3).Caption = CPU_RegDec(CPU.SP)
  
  ' flags
  .selFlag(0).Value = IIf((CPU.PS.N = True), 1, 0)
  .selFlag(1).Value = IIf((CPU.PS.Z = True), 1, 0)
  .selFlag(2).Value = IIf((CPU.PS.V = True), 1, 0)
  .selFlag(3).Value = IIf((CPU.PS.C = True), 1, 0)
  .selFlag(4).Value = IIf((CPU.PS.B = True), 1, 0)
  .selFlag(5).Value = IIf((CPU.PS.D = True), 1, 0)
  .selFlag(6).Value = IIf((CPU.PS.I = True), 1, 0)

  ' NMI?
  .selNMI.Value = IIf((CPU.NMI = True), 1, 0)
  
#If DEMON = 0 Then
  ' instruction history
  MyTxt = ""
  For Recurse = 0 To 7
    MyBase = ExecBase + Recurse
    If (MyBase > 8) Then MyBase = MyBase - 8
    
    If (ExecStack(MyBase) > -1) Then
      MyTxt = MyTxt & "&" & Right$("0000" & _
        Hex(ExecStack(MyBase)), 4) & "  " & _
        CPU_Diss(ExecStack(MyBase)) & vbCrLf
    End If
  Next
  .txtInstHist.Text = MyTxt
  
  ' jump history
  MyTxt = ""
  For Recurse = 0 To 7
    MyBase = JumpBase + Recurse
    If (MyBase > 8) Then MyBase = MyBase - 8
    
    If (JumpStack(MyBase, 1) > -1) Then
      Select Case JumpStack(MyBase, 1)
        Case NMIVEC: MyTxt = MyTxt & " NMIV  --> &" & _
                     Right$("0000" & Hex$(JumpStack(MyBase, 2)), _
                     4) & vbCrLf
                     
        Case IRQVEC: MyTxt = MyTxt & " IRQV  --> &" & _
                     Right$("0000" & Hex$(JumpStack(MyBase, 2)), _
                     4) & vbCrLf
      
        ' there is no RSTVEC as reset clears the stack...
    
        Case Else:
          ' Get the opcode
          OpCode = Memory_Read(JumpStack(MyBase, 1))
          MyTxt = MyTxt & "&" & Right$("0000" & Hex$(JumpStack(MyBase, 1)), 4)
          MyTxt = MyTxt & "  " & Mid$(DisTable, ((OpCode * 3) + 1), 3)
          If ((MyTxt = "BBR") Or (MyTxt = "BBS")) Then
            ' suffix a digit
            OpCode = ((OpCode \ 16) And 7)
            MyTxt = MyTxt & CStr(OpCode) & "--> "
          Else
            MyTxt = MyTxt & " --> "
          End If
          MyTxt = MyTxt & "&" & Right$("0000" & _
                  Hex$(JumpStack(MyBase, 2)), 4) & vbCrLf
      End Select
    End If
  Next
  .txtJumpHist.Text = MyTxt
#End If

  .btnDiss.Caption = "Disassemble from &" & Right$("0000" & Hex$(CPU.PrevPC), 4)
  .btnMemDump.Caption = "Memory dump from &" & Right$("0000" & Hex$(CPU.PrevPC), 4)

  If (ExecHalt = True) Then
    .btnStartStop.Caption = "Run [^Step]"
  Else
    .btnStartStop.Caption = "Halt"
  End If

  EmTime = Cycles / 2000000
  MyTxt = "65C102 status - PC = &" & Right$("0000" & _
          Hex(CPU.PrevPC), 4) & " - " & _
          Format(Cycles, "###,###,###,##0") & " cycles (" & _
          Format(EmTime, "0.##") & "s emulated time)"
  .Caption = MyTxt

  End With
End Sub


Private Function CPU_RegDec(ByVal Reg As Long) As String
  Dim TxtOut As String, Binary As String, Shift As Integer
  
  TxtOut = CStr(Reg) & ", '"
  If ((Reg < 31) Or (Reg = 127) Or ((Reg > 128) And (Reg < 160))) Then
    TxtOut = TxtOut & "[" & Right$("00" & Hex$(Reg), 2) & "]', "
  Else
    TxtOut = TxtOut & Chr$(Reg) & "', "
  End If
  
  For Shift = 1 To 8
    If ((Reg And 1) = 1) Then
      Binary = "1" & Binary
    Else
      Binary = "0" & Binary
    End If
    Reg = Reg \ 2
  Next
  TxtOut = TxtOut & Binary
  
  CPU_RegDec = TxtOut
End Function


#If DEMON = 0 Then
Private Function CPU_Diss(ByVal Where As Long) As String
  ' Simple disassembler
  Dim OpCode As Byte, OpCopy As Byte, TxtOut As String
  Dim MyTemp As Long
  
  ' Get the opcode
  OpCode = Memory_Read(Where)
  TxtOut = Mid$(DisTable, ((OpCode * 3) + 1), 3)
  If ((TxtOut = "RMB") Or (TxtOut = "SMB") Or _
      (TxtOut = "BBR") Or (TxtOut = "BBS")) Then
    ' suffix a digit
    OpCopy = ((OpCode \ 16) And 7)
    TxtOut = TxtOut & CStr(OpCopy) & " "
  Else
    ' suffix two spaces
    TxtOut = TxtOut & "  "
  End If
  
  ' Now work out what to display
  Select Case AddrModes(OpCode)
    Case AXXX: TxtOut = TxtOut & "##INVALID##"
        
    Case AACC: TxtOut = TxtOut & "A"

    Case AIMM: MyTemp = Memory_Read(Where + 1&)
               TxtOut = TxtOut & "#&" & Hex$(MyTemp)
      
    Case AABS: MyTemp = CPU_ReadAddress(Where + 1&)
               TxtOut = TxtOut & "&" & Right$("0000" & _
                        Hex$(MyTemp), 4)
    
    Case AZPA: MyTemp = Memory_Read(Where + 1&)
               TxtOut = TxtOut & "&" & Right$("00" & _
                        Hex$(MyTemp), 2)
    
    Case AZPX: MyTemp = Memory_Read(Where + 1&)
               TxtOut = TxtOut & "&" & Right$("00" & _
                        Hex$(MyTemp), 2) & ",X"

    Case AZPY: MyTemp = Memory_Read(Where + 1&)
               TxtOut = TxtOut & "&" & Right$("00" & _
                        Hex$(MyTemp), 2) & ",Y"
    
    Case AABX: MyTemp = CPU_ReadAddress(Where + 1&)
               TxtOut = TxtOut & "&" & Right$("0000" & _
                        Hex$(MyTemp), 4) & ",X"
    
    Case AABY: MyTemp = CPU_ReadAddress(Where + 1&)
               TxtOut = TxtOut & "&" & Right$("0000" & _
                        Hex$(MyTemp), 4) & ",Y"
      
    Case AIMP: ' nothing to add
    
    Case AREL: MyTemp = Memory_Read(Where + 1&)
               If (MyTemp > 127) Then MyTemp = (-256& + MyTemp)
               MyTemp = Where + MyTemp + 2
               TxtOut = TxtOut & "&" & Right$("0000" & _
                        Hex$(MyTemp), 4)

    Case AINX: MyTemp = Memory_Read(Where + 1)
               TxtOut = TxtOut & "(&" & Right$("00" & _
                        Hex$(MyTemp), 2) & ",X)"
      
    Case AINY: MyTemp = Memory_Read(Where + 1)
               TxtOut = TxtOut & "(&" & Right$("00" & _
                        Hex$(MyTemp), 2) & "),Y"
    
    Case AABI: MyTemp = CPU_ReadAddress(Where + 1)
               TxtOut = TxtOut & "(&" & Right$("0000" & _
                        Hex$(MyTemp), 4) & ")"
    
    Case AIAI: MyTemp = Memory_Read(Where + 1)
               TxtOut = TxtOut & "(&" & Right$("0000" & _
                        Hex$(MyTemp), 4) & ",X)"

    Case AZPR: MyTemp = Memory_Read(Where + 1)
               TxtOut = TxtOut & "&" & Right$("00" & _
                        Hex$(MyTemp), 2)
                        
    
    Case AIND: MyTemp = Memory_Read(Where + 1)
               TxtOut = TxtOut & "(&" & Right$("00" & _
                        Hex$(MyTemp), 2) & ")"
    
    Case Else: TxtOut = TxtOut & "?????"
  End Select

  CPU_Diss = TxtOut
End Function
#End If


Public Function CPU_DissFancy(ByVal Where As Long, DrawOn As Object, _
                              ByVal LHei As Long) As Integer
  ' More detailed part-colourised disassembler; writes to 'DrawOn'.
  ' Expects DrawOn to have been set up with 10pt Courier New.
  ' Returns instruction bytecount for updating local copy of PC.
  Dim OpCode As Byte, OpCopy As Byte, TxtOut As String, MyByte As Byte
  Dim MyTemp As Long, ByteCnt As Integer, Recurse As Integer, ThisAnno As String
  Dim OldX As Long, OldY As Long, Inverse As Boolean
  
  ' The current instruction?
  If (Where = CPU_ReadPC) Then
    OldX = DrawOn.CurrentX
    OldY = DrawOn.CurrentY
    DrawOn.Line -(DrawOn.Width, OldY + LHei), vbBlack, BF
    DrawOn.CurrentX = OldX
    DrawOn.CurrentY = OldY
    DrawOn.ForeColor = vbWhite
    Inverse = True
  Else
    DrawOn.ForeColor = vbBlack
    Inverse = False
  End If
  
  ' Write the address
  DrawOn.Font.Bold = False
  DrawOn.Font.Italic = False
  DrawOn.Print "&" & Right$("0000" & Hex$(Where), 4) & " : ";
  
  ' Get the opcode
  OpCode = Memory_Read(Where)
  TxtOut = Mid$(DisTable, ((OpCode * 3) + 1), 3)
  If ((TxtOut = "RMB") Or (TxtOut = "SMB") Or _
      (TxtOut = "BBR") Or (TxtOut = "BBS")) Then
    ' suffix a digit
    OpCopy = ((OpCode \ 16) And 7)
    TxtOut = TxtOut & CStr(OpCopy) & " "
  Else
    ' suffix two spaces
    TxtOut = TxtOut & "  "
  End If
  
  ' Now work out what to display
  Select Case AddrModes(OpCode)
    Case AXXX: TxtOut = TxtOut & "##INVALID##"
               ByteCnt = 1
        
    Case AACC: TxtOut = TxtOut & "A"
               ByteCnt = 1

    Case AIMM: MyTemp = Memory_Read(Where + 1&)
               TxtOut = TxtOut & "#&" & Hex$(MyTemp)
               ByteCnt = 2
      
    Case AABS: MyTemp = CPU_ReadAddress(Where + 1&)
               TxtOut = TxtOut & "&" & Right$("0000" & _
                        Hex$(MyTemp), 4)
               ByteCnt = 3
    
    Case AZPA: MyTemp = Memory_Read(Where + 1&)
               TxtOut = TxtOut & "&" & Right$("00" & _
                        Hex$(MyTemp), 2)
               ByteCnt = 2
    
    Case AZPX: MyTemp = Memory_Read(Where + 1&)
               TxtOut = TxtOut & "&" & Right$("00" & _
                        Hex$(MyTemp), 2) & ",X"
               ByteCnt = 2

    Case AZPY: MyTemp = Memory_Read(Where + 1&)
               TxtOut = TxtOut & "&" & Right$("00" & _
                        Hex$(MyTemp), 2) & ",Y"
               ByteCnt = 2
    
    Case AABX: MyTemp = CPU_ReadAddress(Where + 1&)
               TxtOut = TxtOut & "&" & Right$("0000" & _
                        Hex$(MyTemp), 4) & ",X"
               ByteCnt = 3
    
    Case AABY: MyTemp = CPU_ReadAddress(Where + 1&)
               TxtOut = TxtOut & "&" & Right$("0000" & _
                        Hex$(MyTemp), 4) & ",Y"
               ByteCnt = 3
      
    Case AIMP: ' nothing to add
               ByteCnt = 1
  
    Case AREL: MyTemp = Memory_Read(Where + 1&)
               If (MyTemp > 127) Then MyTemp = (-256& + MyTemp)
               MyTemp = Where + MyTemp + 2
               TxtOut = TxtOut & "&" & Right$("0000" & _
                        Hex$(MyTemp), 4)
               ByteCnt = 2

    Case AINX: MyTemp = Memory_Read(Where + 1)
               TxtOut = TxtOut & "(&" & Right$("00" & _
                        Hex$(MyTemp), 2) & ",X)"
               ByteCnt = 2
      
    Case AINY: MyTemp = Memory_Read(Where + 1)
               TxtOut = TxtOut & "(&" & Right$("00" & _
                        Hex$(MyTemp), 2) & "),Y"
               ByteCnt = 2
    
    Case AABI: MyTemp = CPU_ReadAddress(Where + 1)
               TxtOut = TxtOut & "(&" & Right$("0000" & _
                        Hex$(MyTemp), 4) & ")"
               ByteCnt = 3
    
    Case AIAI: MyTemp = Memory_Read(Where + 1)
               TxtOut = TxtOut & "(&" & Right$("0000" & _
                        Hex$(MyTemp), 4) & ",X)"
               ByteCnt = 2

    Case AZPR: MyTemp = Memory_Read(Where + 1)
               TxtOut = TxtOut & "&" & Right$("00" & _
                        Hex$(MyTemp), 2)
               ByteCnt = 2
                            
    Case AIND: MyTemp = Memory_Read(Where + 1)
               TxtOut = TxtOut & "(&" & Right$("00" & _
                        Hex$(MyTemp), 2) & ")"
               ByteCnt = 2
    
    Case Else: TxtOut = TxtOut & "?????"
               ByteCnt = 1
  End Select
  
  ' Now output the bytes of this instruction in hex
  For Recurse = 0 To (ByteCnt - 1)
    DrawOn.Print Right$("00" & Hex$(Memory_Read(Where + Recurse)), 2) & " ";
  Next
  If (ByteCnt < 3) Then
    DrawOn.Print Space$((3 - ByteCnt) * 3);
  End If
  DrawOn.Print " : ";
  
  ' Now output the bytes of this instruction in hex
  For Recurse = 0 To (ByteCnt - 1)
    MyByte = Memory_Read(Where + Recurse)
    Select Case MyByte
      Case 0 To 32, 127, 129, 141, 143, 144, 157: MyByte = 46
    End Select
    DrawOn.Print Chr$(MyByte);
  Next
  If (ByteCnt < 3) Then
    DrawOn.Print Space$(3 - ByteCnt);
  End If
  DrawOn.Print " : ";
  DrawOn.Font.Bold = True
  DrawOn.Print TxtOut;
  DrawOn.Font.Bold = False
  
  ThisAnno = Memory_LookupAnno(Where)
  If (Len(ThisAnno) <> 0) Then
    DrawOn.Print Space$(20 - Len(TxtOut));
    If (Inverse) Then
      DrawOn.ForeColor = &HD000&
    Else
      DrawOn.ForeColor = &H8000&
    End If
    DrawOn.Font.Bold = True
    DrawOn.Print "; == ";
    DrawOn.Font.Bold = False
    DrawOn.Font.Italic = True
    DrawOn.Print ThisAnno;
    DrawOn.ForeColor = vbBlack
    DrawOn.Font.Italic = False
    TxtOut = "!ANNO!"
  End If
  
  ThisAnno = Memory_LookupAnno(MyTemp)
  If (Len(ThisAnno) <> 0) Then
    If (Inverse) Then
      DrawOn.ForeColor = &HD000&
    Else
      DrawOn.ForeColor = &H8000&
    End If
    If (TxtOut <> "!ANNO!") Then
      DrawOn.Print Space$(20 - Len(TxtOut));
    Else
      DrawOn.Print ", ";
    End If
    DrawOn.Font.Bold = True
    DrawOn.Print "; -> ";
    DrawOn.Font.Bold = False
    DrawOn.Font.Italic = True
    DrawOn.Print ThisAnno;
    DrawOn.ForeColor = vbBlack
    DrawOn.Font.Italic = False
  End If
  
  ' done
  DrawOn.Print

  CPU_DissFancy = ByteCnt
End Function


Public Sub CPU_DissUpdatePopup()
  ' Updates the disassembly popup
  Dim MyTxt As String
  
  frmDiss.picPopup.Cls
  MyTxt = "   X = &" & Right$("00" & Hex$(CPU.X), 2) & _
          "   Y = &" & Right$("00" & Hex$(CPU.Y), 2) & _
          "   A = &" & Right$("00" & Hex$(CPU.A), 2) & vbCrLf & _
          "  SP = &" & Right$("00" & Hex$(CPU.SP), 2) & _
          "  PC = &" & Right$("0000" & Hex$(CPU.PrevPC), 4) & vbCrLf
  frmDiss.picPopup.Print MyTxt
  
  ' NV-DBIZC (order as given in Beeb tech ref manual)
  If (CPU.PS.N = True) Then
    MyTxt = "  PS = N"
  Else
    MyTxt = "  PS = n"
  End If
  If (CPU.PS.V = True) Then
    MyTxt = MyTxt & "V-"
  Else
    MyTxt = MyTxt & "v-"
  End If
  If (CPU.PS.B = True) Then
    MyTxt = MyTxt & "B"
  Else
    MyTxt = MyTxt & "b"
  End If
  If (CPU.PS.D = True) Then
    MyTxt = MyTxt & "D"
  Else
    MyTxt = MyTxt & "d"
  End If
  If (CPU.PS.I = True) Then
    MyTxt = MyTxt & "I"
  Else
    MyTxt = MyTxt & "i"
  End If
  If (CPU.PS.Z = True) Then
    MyTxt = MyTxt & "Z"
  Else
    MyTxt = MyTxt & "z"
  End If
  If (CPU.PS.C = True) Then
    MyTxt = MyTxt & "C"
  Else
    MyTxt = MyTxt & "c"
  End If
  If (CPU.NMI = True) Then MyTxt = MyTxt & "  [NMI]"
  frmDiss.picPopup.Print MyTxt
End Sub


Private Sub CPU_Fiddle(ByVal Op As Byte, _
                       ByVal Addr As Long, _
                       ByVal Why As AbortReason)
  ' Allows some degree of recoverability...
  
  Dim MyMsg As String

  Select Case Why
    Case INSTRUNKNOWN:
      MyMsg = "Unrecognised instruction &" & Right$("00" & _
              Hex(Op), 2) & " at address &" & _
              Right$("0000" & Hex(Addr), 4) & "."
    
    Case ADDRUNKNOWN:
      MyMsg = "The instruction &" & Right$("00" & _
              Hex(Op), 2) & " at address &" & _
              Right$("0000" & Hex(Addr), 4) & " is using " & _
              "an unrecognised addressing mode."
  
    Case INSTRUNIMPL:
      MyMsg = "The instruction &" & Right$("00" & _
              Hex(Op), 2) & " at address &" & _
              Right$("0000" & Hex(Addr), 4) & " has not " & _
              "been implemented."
  End Select
  
  MyMsg = MyMsg & vbCrLf & vbCrLf & "You may be able to fake " & _
    "the behaviour of this instruction by editing the CPU " & _
    "state; however for a more solid correction, you should " & _
    "edit the ""CPU65C102.bas"" source to do this as a matter " & _
    "of course..."
  
  MsgBox MyMsg, vbExclamation + vbOKOnly
  
  CPU.Abort = True
  
  frmCPUState.Show
  CPU_RefreshTrace ' only works when window is 'visible'
  frmCPUState.Hide
  
  frmCPUState.Show

End Sub


Public Sub CPU_SetFlag(ByVal Which As Integer, Value As Boolean)
  ' N Z V C  B D I  NMI

  Select Case Which
    Case 0: CPU.PS.N = Value
    Case 1: CPU.PS.Z = Value
    Case 2: CPU.PS.V = Value
    Case 3: CPU.PS.C = Value
    Case 4: CPU.PS.B = Value
    Case 5: CPU.PS.D = Value
    Case 6: CPU.PS.I = Value
    
    Case 7:
      If ((Value = True) And (CPU.NMI = False)) Then
        If (MsgBox("Are you sure you wish to set NMI lockout?", _
                   vbExclamation + vbYesNo + vbDefaultButton2) _
          = vbYes) Then CPU.NMI = True
      Else
        CPU.NMI = Value
      End If
  End Select
End Sub


Public Sub CPU_SetReg(ByVal Which As Integer, ByVal Value As String)
  ' X  Y  A  SP
  
  Dim MyVal As Byte
  
  MyVal = (Val("&H" & Value) And 255)
  
  Select Case Which
    Case 0: CPU.X = MyVal
    Case 1: CPU.Y = MyVal
    Case 2: CPU.A = MyVal
    
    Case 3:
        If (MsgBox("Are you sure you wish to change the Stack Pointer?", _
                   vbExclamation + vbYesNo + vbDefaultButton2) _
          = vbYes) Then CPU.SP = MyVal
          
    Case -3: CPU.SP = MyVal ' for testscript
  End Select
End Sub


Public Function CPU_ReadReg(ByVal Which As Integer) As Byte
  ' X  Y  A  SP  [& for testscript: PC PS C Z I D B V N]
  Select Case Which
    Case 0: CPU_ReadReg = CPU.X
    Case 1: CPU_ReadReg = CPU.Y
    Case 2: CPU_ReadReg = CPU.A
    Case 3: CPU_ReadReg = CPU.SP
    
    ' for testscript...
    Case 4: ' CPU_ReadReg = CPU.PC  << overflows! :-)
    Case 5: ' CPU_ReadReg = CPU.PS  << need a C-style union here
    Case 6: CPU_ReadReg = CPU.PS.C
    Case 7: CPU_ReadReg = CPU.PS.Z
    Case 8: CPU_ReadReg = CPU.PS.I
    Case 9: CPU_ReadReg = CPU.PS.D
    Case 10: CPU_ReadReg = CPU.PS.B
    Case 11: CPU_ReadReg = CPU.PS.V
    Case 12: CPU_ReadReg = CPU.PS.N
  End Select
End Function


Public Function CPU_Abort() As Long
  ' Sets the abort flag to drop out of executing...  [why is this a FUNCTION???]
  CPU.Abort = True
End Function


Public Sub CPU_PrepForTest()
  ' Sets single-step status, and PC to &200.
  ExecHalt = True
  CPU.PC = &H200&
End Sub

