Attribute VB_Name = "Memory"
' RAM, EPROM, and memory decode/dispatch.
' Also contains some memory-related routines,
' such as code for the EPROM image joining.
'
' Project: FileStore [E01S emulator]
' Module : Memory
'
' Version: 0.00
' Date   : 2nd August 2009
' Started: 11th April 2009
'
' Author : Rick Murray, heyrick1973 -at- yahoo -dot- co -dot- uk
'          http://www.heyrick.co.uk/econet/
'
'
' Licence:
'
'   Copyright 2010 Rick Murray
'
'   Licensed under the EUPL, Version 1.1 ONLY (the "Licence");
'   You may not use this work except in compliance with the Licence.
'
'   You may obtain a copy of the Licence at:
'
'     http://ec.europa.eu/idabc/eupl5
'
'   Unless required by applicable law or agreed to in writing, software
'   distributed under the Licence is distributed on an "AS IS" basis,
'   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
'   implied. See the Licence for the specific language governing
'   permissions and limitations under the Licence.
'
'
'   Rick's note: The EUPL is "compatible" with GPL v2, but *NOT* the more
'                restrictive GPL v3. If you will be involving GPL code
'                in this project, or using some of this code in a GPL
'                project, you should specify the use of ONLY the GPL v2
'                (and not the boilerplate "GPL 2 or any later...").
'

Option Explicit

Private MemEPROMLatch As Boolean   ' EPROM 'visible'?
Private RAM(0 To 65535) As Byte    ' RAM
Private ROM(0 To 65535) As Byte    ' EPROM space.
Private ROMCHK As Long             ' Used for our patch code.

Private Type DissAnnoDef
  Addr As Long
  Anno As String
End Type
Private DissAnno() As DissAnnoDef
Private DissAnnoCnt As Integer

Private Const MRAM As Integer = 1  ' RAM
Private Const MROM As Integer = 2  ' EPROM
Private Const MRTC As Integer = 3  ' Real Time Clock
Private Const MMSG As Integer = 4  ' Memory Selection Gate
Private Const MMFL As Integer = 5  ' Misc. Functions Latch
Private Const MFDC As Integer = 6  ' Floppy Disc Controller
Private Const MVIA As Integer = 7  ' Versatile Interface Adaptor
Private Const MNET As Integer = 8  ' NETwork ADLC
Private Const MNIC As Integer = 9  ' Network Interrupt Control
Private Const MFPS As Integer = 10 ' Front Panel State
Private Const MHDC As Integer = 11 ' Hard Disc Controller
Private Const MUHA As Integer = 12 ' Unknown Hardware Access


Public Sub Memory_Initialise()
  ' This will load the EPROM image file into the EPROM array.
  
  Dim Recurse As Long, MyByte As Byte, FP As Integer
  Dim MyFile As String, Cnt As Integer
  
  ' Check the image file exists
  MyFile = Setup_RebuildFile(SetupInfo.ROMImage)
  If (Len(Dir(MyFile)) = 0) Then
    MsgBox "Unable to find the EPROM image file """ & _
           MyFile & """.", vbExclamation
    Shutdown
  End If
  
  ROMCHK = 0
  
  ' Load the image
  FP = FreeFile
  Open MyFile For Binary Access Read As FP
  If (LOF(FP) <> 65536) Then
    MsgBox "The EPROM image file is not 64KiB!" & vbCrLf & _
           vbCrLf & "If you have the image in smaller parts " & _
           "(i.e. 416KiB or 232KiB) you can use the """ & _
           "Image Joiner"" to create the desired 64KiB image.", _
           vbExclamation + vbOKOnly
    Shutdown
  End If
  frmInit.lblStatus = "Initialising; loading EPROM image..."
  frmInit.lblStatus.Refresh
  For Recurse = 0& To 65535
    Get #FP, , MyByte
    ROM(Recurse) = MyByte
    ROMCHK = ((ROMCHK + CLng(MyByte)) And &HFFFF&)
  Next
  Close #FP
  
  
  ' look for, and load if available, code offset annotations
  DissAnnoCnt = 0
  ReDim DissAnno(DissAnnoCnt)
  MyFile = App.Path & "\locs" & Right$("0000" & Hex$(ROMCHK), 4) & ".dat"
  If (Len(Dir(MyFile)) = 0) Then
    ' try the generic annotations file
    MyFile = App.Path & "\locsxxxx.dat"
  End If
  If (Len(Dir(MyFile)) <> 0) Then
    FP = FreeFile
    Open MyFile For Binary Access Read As FP
    Cnt = 0
    Do While (Loc(FP) < LOF(FP))
      Line Input #FP, MyFile
      Cnt = Cnt + 1
    Loop
    Seek #FP, 1
    DissAnnoCnt = Cnt
    ReDim DissAnno(DissAnnoCnt)
    Cnt = 1
    Do While (Loc(FP) < LOF(FP))
      Line Input #FP, MyFile
      If (Mid$(MyFile, 5, 1) = ":") Then
        DissAnno(Cnt).Addr = CLng("&H" & Left$(MyFile, 4))
        DissAnno(Cnt).Anno = Mid$(MyFile, 6)
        
        Cnt = Cnt + 1
      End If
    Loop
    Close #FP
  End If
  
  ' Clear the RAM - this is not 'real' but it will aid in
  ' examining the memory if the need should arise...
  frmInit.lblStatus = "Initialising; clearing RAM..."
  frmInit.lblStatus.Refresh
  For Recurse = 0& To 65535
    RAM(Recurse) = 0
  Next
  
  ' Set the EPROM latch to be active so we can 'see' the EPROM
  MemEPROMLatch = True
End Sub


Public Function Memory_Read(ByVal Addr As Long) As Byte
  ' Reads a byte from memory
  
  Dim Locn As Integer, Reply As Byte, WrkSpc As String
  
  Locn = Memory_Lookup(Addr, True) ' work out where/what/how
  
  Select Case Locn
    Case MRAM: Reply = RAM(Addr)
    
    Case MROM: Reply = ROM(Addr)
    
    Case MRTC: Reply = NVRAM_Read(CInt(Addr And 7&))
    
    Case MMSG: ' *READING* this address will disable EPROM
               Reply = 0
               MemEPROMLatch = False
    
    '    MMFL: ' Misc. Functions Latch is WRITE ONLY.
    
    Case MFDC: Reply = FDC_Read(CInt(Addr - &HFC0C&) And 3)
    
    Case MVIA: Reply = VIA_Read(CInt(Addr - &HFC10&) And 15)
    
    Case MNET: Reply = ADLC_Read(CInt(Addr - &HFC20&) And 3)
    
    Case MNIC: ' does a READ have the same effect?
               ' we'll assume YES, correct me if this is not so.
               If (Addr < &HFC28&) Then
                 ' &FC24 - &FC27, disable Econet interrupts
                 ADLCmod.CanIRQ = False
                 HWLog "Network IRQs disabled [READ]"
               
               Else
                 ' &FC28 - &FC2B, enable Econet interrupts
                 ADLCmod.CanIRQ = True
                 HWLog "Network IRQs enabled [READ]"
               End If
    
    Case MFPS: ' call front panel state
               ' do it here, it's easy...
               WrkSpc = "Reading front panel state (is "
               If (FlapOpen = True) Then
                 Reply = 64 ' bit 6 set
                 WrkSpc = WrkSpc & "open)"
               Else
                 Reply = 0
                 WrkSpc = WrkSpc & "closed)"
               End If
               HWLog WrkSpc
               ' ##TODO## if we want to put in a complete
               ' emulation of the E01 FileStore, the internal
               ' switch "SW3" is wired to bit 7; but was this
               ' ever actually used for anything!?
    
    Case MHDC: Reply = HDC_Read(CInt(Addr - &HFC30&) And 3)
    
    Case Else: MsgBox "Unknown memory READ access - address &" & _
                      Right$("0000" & Hex(Addr), 4), vbExclamation
               Shutdown ' ## CURRENTLY A FATAL ERROR ##
  End Select
  
  Memory_Read = Reply
End Function


Public Sub Memory_ReadPage(ByRef PageBuff() As Byte, _
                           ByVal Page As Long)
  ' For the memory display. This will directly fill a buffer with
  ' the contents of the page specified. It only reads from RAM and
  ' it ignores memory decoding - so this is only suitable for use
  ' with the memory dump; and also since page &FC should be masked
  ' out...
  
  Dim Recurse As Long
  
  For Recurse = 0 To 255
    PageBuff(Recurse) = RAM((Page * 256&) + Recurse)
  Next
End Sub


Public Sub Memory_Write(ByVal Value As Byte, ByVal Addr As Long)
  ' Writes a byte to memory
  
  Dim Locn As Integer
  
  Locn = Memory_Lookup(Addr, False) ' who/what/where/why/how?
  
  Select Case Locn
    Case MRAM: RAM(Addr) = Value
    
    Case MROM: MsgBox "Attempted to WRITE to the EPROM!", _
                      vbExclamation
               Shutdown ' ## CURRENTLY A FATAL ERROR ##
    
    Case MRTC: NVRAM_Write Value, CInt((Addr And 7&))
    
    '    MMSG: ' EPROM latch is READ ONLY (write is MiscFuncLatch)
    
    Case MMFL: Latch_Write Value
    
    Case MFDC: FDC_Write Value, (CInt(Addr - &HFC0C&) And 3)
    
    Case MVIA: VIA_Write Value, (CInt(Addr - &HFC10&) And 15)
    
    Case MNET: ADLC_Write Value, (CInt(Addr - &HFC20&) And 3)
    
    Case MNIC: ' does a READ have the same effect?
               ' we'll assume YES, correct me if this is not so.
               If (Addr < &HFC28&) Then
                 ' &FC24 - &FC27, disable Econet interrupts
                 ADLCmod.CanIRQ = False
                 HWLog "Network IRQs disabled [WRITE]"
               
               Else
                 ' &FC28 - &FC2B, enable Econet interrupts
                 ADLCmod.CanIRQ = True
                 HWLog "Network IRQs enabled [WRITE]"
               End If
    
    '    MFPS: ' Front panel state is READ ONLY
    
    Case MHDC: HDC_Write Value, (CInt(Addr - &HFC30&) And 3)
    
    Case Else: MsgBox "Unknown memory WRITE access - address &" & _
                      Right$("0000" & Hex(Addr), 4), vbExclamation
               Shutdown ' ## CURRENTLY A FATAL ERROR ##
  End Select
  
End Sub


Public Sub Memory_Poke(ByVal RAMAddr As Long, ByVal Value As Byte)
  ' Writes a byte to memory, RAM array. Parameters are other way
  ' around to Memory_Write to make it (hopefully) just a little
  ' bit harder to mix 'em up.

  RAM(RAMAddr) = Value
End Sub


Private Function Memory_Lookup(ByVal Addr As Long, _
                               ByVal Read As Boolean) _
                               As Integer
  ' We don't do a complete decode here; we just know which bit
  ' of hardware we are looking to read/write. For example we
  ' do not currently CARE which VIA register, we just want to
  ' know it's the VIA...
  
  Dim HWAddr As Integer
  
  ' Hardware?
  If (Addr >= &HFC00&) And (Addr <= &HFCFF&) Then
    ' It is an access into memory-mapped hardware
    '
    ' We support the modulo &40 addressing scheme, plus the
    ' alternate addresses of devices.
    ' It is, however, STRONGLY recommended that all code
    ' refers to the legal addresses for obvious reasons...
    
    HWAddr = CInt(Addr And &H3F&)
        
    Memory_Lookup = MUHA ' default to 'unknown' (until otherwise)
    
    If ((HWAddr >= &H0) And (HWAddr <= &H7)) Then
      Memory_Lookup = MRTC   ' NVRAM &FC00 = Addr, &FC04 = Data
                             ' alts: ...&FC03      ...&FC07
      
    ElseIf ((HWAddr >= &H8) And (HWAddr <= &HB)) Then
      If (Read = True) Then
        Memory_Lookup = MMSG ' &FC08 READ is EPROM/RAM selector
      Else                   '   alt: ...&FC0B
        Memory_Lookup = MMFL ' and WRITE is the Misc. Funky Latch
      End If
    
    ElseIf ((HWAddr >= &HC) And (HWAddr <= &HF)) Then
      Memory_Lookup = MFDC   ' &FC0C-&FC0F is floppy disc; no alt
    
    ElseIf ((HWAddr >= &H10) And (HWAddr <= &H1F)) Then
      Memory_Lookup = MVIA   ' &FC1x is the VIA; no alt
    
    ElseIf ((HWAddr >= &H20) And (HWAddr <= &H23)) Then
      Memory_Lookup = MNET   ' &FC20-&FC23 is the ADLC; no alt

    ElseIf ((HWAddr >= &H24) And (HWAddr <= &H2B)) Then
      Memory_Lookup = MNIC   ' &FC24 = No net IRQ, &FC28 = Enable
                             ' alts: ...&FC27      ...&FC0B

    ElseIf ((HWAddr >= &H2C) And (HWAddr <= &H2F)) Then
      Memory_Lookup = MFPS   ' &FC2C is the front panel sw (bit6)
                             ' alt: ...&FC2F   / E01=SW3 on bit7
      
    ElseIf (HWAddr >= &H30) And (HWAddr <= &H33) Then
      Memory_Lookup = MHDC   ' &FC30-&FC33 is the SCSI hardware
    End If                   ' no alts?

  ' Not hardware, so it must be either EPROM or RAM
  ElseIf (MemEPROMLatch = True) Then
    If (Read = True) Then
      Memory_Lookup = MROM ' read from EPROM
    Else
      Memory_Lookup = MRAM ' and write to RAM
    End If
    Exit Function
  
  Else
    Memory_Lookup = MRAM   ' always RAM access if latch clear
  End If
End Function


Public Sub Memory_PatchROMOLD()
  ' What we do is PATCH the MOS code to not make us sit
  ' through an annoyingly long startup.
  ' The ~10s delay probably makes sense on REAL hardware,
  ' where we might need to wait for drives to spin up and
  ' the like. But for an emulation?...
  '
  ' We patch the EPROM image in MEMORY, the version on file
  ' is NOT altered.
  
  Dim BaseAddr As Long, MyByte As Byte, MyChk As Integer
  Dim Recurse As Integer
  
  BaseAddr = 0&
  
  frmInit.lblStatus = "Initialising; patching ROM..."
  frmInit.lblStatus.Refresh
  
  Select Case ROMCHK
    Case &H31F0&:
      ' E01 firmware "FileStore 1.31"       (C)Acorn 1988
      BaseAddr = &HECA5&
    
    Case &HFE41&:
      ' E01S firmware "FileStore 1.33"      (C)Acorn 1988
      BaseAddr = &HECA9&
      
    Case &HE529&:
      ' E01S firmware "FileStore 1.40.00"   (C)1989 Acorn
      BaseAddr = &HECAF&
    
    ' ##TODO## Add recognition for other versions here.
  End Select
  
  ' What lies at these addresses should be:
  '
  '    ECA5  AD 68 E7  LDA  &E768
  '    ECA8  C9 04     CMP  #&04
  '    ECAA  90 F9     BCC  &ECA5 (or whatever is -7 bytes...)
  '
  ' which reads the second byte of the internal ticker until it
  ' reads '4' [(4 * 256) / 100] which should be roughly 10 secs.
  ' (or somewhat longer in emulation time!)
  
  ' Verify this is the correct code (just add up the bytes)
  MyChk = 0
  For Recurse = 0 To 6
    MyChk = MyChk + ROM(BaseAddr + CLng(Recurse))
  Next
  If (MyChk <> 1106) Then Exit Sub
  
  ' Now write NOPs so we'll just skip on to the FileServer startup
  For Recurse = 0 To 6
    ROM(BaseAddr + CLng(Recurse)) = &HEA
  Next
End Sub


Public Sub Memory_PatchROM()
  ' What we do is PATCH the MOS code to not make us sit
  ' through an annoyingly long startup.
  ' The ~10s delay probably makes sense on REAL hardware,
  ' where we might need to wait for drives to spin up and
  ' the like. But for an emulation?...
  '
  ' We patch the EPROM image in MEMORY, the version on file
  ' is NOT altered.
  
  Dim BaseAddr As Long, MyByte As Byte, MyChk As Integer, BLen As Integer
  Dim Recurse As Integer, FP As Integer, Patch As String, PLin As Integer
  Dim ChkVal As Integer
  
  ' We have a patch file, right?
  If (Len(Dir(App.Path & "\rompatch.dat")) = 0) Then Exit Sub
  
  ' Open the file
  FP = FreeFile
  Open App.Path & "\rompatch.dat" For Binary Access Read As FP
  
  ' Read through the file looking for applicable matches
  Do While (Loc(FP) < LOF(FP))
    Line Input #FP, Patch
    PLin = PLin + 1
    
    ' clip off from the first semicolon found
    BLen = InStr(1, Patch, ";")
    If (BLen = 1) Then
      Patch = ""
    Else
      Patch = Left$(Patch, (BLen - 1))
    End If
    
    If (Patch <> "") Then
      BaseAddr = (Val("&H" & Left$(Patch, 4)) And &HFFFF&)
    
      If (BaseAddr = ROMCHK) Then
        ' Okay, this is for us...
        If (Mid$(Patch, 5, 1) <> "@") Then
          Close #FP
          MsgBox "Invalid EPROM patch - '@' missing or misplaced in line " & _
                 CStr(PLin) & "." & vbCrLf & "Aborting patch.", vbExclamation
          Exit Sub
        End If
        
        BaseAddr = (Val("&H" & Mid$(Patch, 6, 4)) And &HFFFF&)
    
        frmInit.lblStatus = "Initialising; patching ROM..."
        frmInit.lblStatus.Refresh
        
        If (Mid$(Patch, 10, 1) <> ">") Then
          Close #FP
          MsgBox "Invalid EPROM patch - '>' missing or misplaced in line " & _
                 CStr(PLin) & "." & vbCrLf & "Aborting patch.", vbExclamation
          Exit Sub
        End If
        
        Patch = Mid$(Patch, 11)
        BLen = (Val(Patch) - 1) ' if is "7:1234", will reply '7'.
        ' we subtract 1 as we may patch 8 bytes, but that means we
        ' count from ZERO to SEVEN.
        
        Recurse = InStr(1, Patch, ":")
        Patch = Mid$(Patch, (Recurse + 1))
        ChkVal = Val("&H" & Left$(Patch, 5))
        
        ' okay, we have the address, the number of bytes, and the check
        ' value... so let's confirm it.
        MyChk = 0
        For Recurse = 0 To BLen
          MyChk = MyChk + ROM(BaseAddr + CLng(Recurse))
        Next
        If (MyChk <> ChkVal) Then
          ' you can use this if you don't have the inclination to work
          ' out the correct check value yourself - simply set the value
          ' in the file to '0000' and wait for this error message to
          ' tell you what the correct value is! :-)
          Close #FP
          MsgBox "Invalid EPROM patch - check value was &" & Right$( _
                 "0000" & Hex$(MyChk), 4) & " and not the value &" & _
                 Right$("0000" & Hex$(ChkVal), 4) & " expected." & _
                 vbCrLf & "Aborting patch.", vbExclamation
          Exit Sub
        End If
        
        ' okay, it all makes sense so...
        For Recurse = 0 To BLen
          MyByte = Val("&H" & Mid$(Patch, ((Recurse * 3) + 1), 2))
          ROM(BaseAddr + CLng(Recurse)) = MyByte
        Next
        
      End If ' our EPROM?
    End If ' not a blank line?
  Loop ' to read a new line...
  
  Close #FP
End Sub


Public Sub Memory_DoJoin(ImgF() As String, _
                         Dest As String, iCnt As Integer)
  Dim OFP As Integer, FN As String, Recurse As Long, MyByte As Byte, Wank As Long

  ' some user feedback; wonk it to fit our specifics...
  Load frmInit
  frmInit.lblTitle.Caption = "EPROM Image Joiner"
  frmInit.lblVersion.Caption = ""
  frmInit.lblURL.Caption = "Joining " & CStr(iCnt) & " parts..."
  frmInit.lblLicence.Caption = ""
  frmInit.lblStatus = ""
  frmInit.Show
  DoEvents

  ' blank memory
  frmInit.lblStatus = "EPROM joiner: clearing workspace..."
  frmInit.lblStatus.Refresh
  For Recurse = 0 To 65535
    ROM(Recurse) = 0
  Next
  
  If (iCnt = 2) Then
    ' 2  32KiB
    Memory_JoinUp ImgF(1), &H0&, &H8000&, 1     ' first 32KiB
    Memory_JoinUp ImgF(2), &H8000&, &H8000&, 2  ' second 32KiB
  
  Else
    ' 4  16KiB
    Memory_JoinUp ImgF(1), &H0&, &H4000&, 1     ' first 16KiB
    Memory_JoinUp ImgF(2), &H4000&, &H4000&, 2  ' second 16KiB
    Memory_JoinUp ImgF(3), &H8000&, &H4000&, 3  ' third 16KiB
    Memory_JoinUp ImgF(4), &HC000&, &H4000&, 4  ' fourth 16KiB
  End If

  ' now write it out as a single dump
  frmInit.lblStatus = "EPROM joiner: writing complete image..."
  frmInit.lblStatus.Refresh
  FN = Setup_RebuildFile(Dest)
  If (Len(Dir(FN)) > 0) Then Kill FN
  OFP = FreeFile
  Open FN For Binary Access Write As OFP
  For Recurse = 0 To 65535
    MyByte = ROM(Recurse)
    Put #OFP, , MyByte
  Next
  Close #OFP

  ' set image info in Setup to match
  frmSetup.txtImgFile(3).Text = Dest
  
  Unload frmInit
  
  MsgBox "Image file assembled.", vbInformation + vbOKOnly
End Sub
                         

Private Sub Memory_JoinUp(FN As String, Base As Long, _
                          FLen As Long, Cnt As Integer)
  ' Copies <FLen> bytes from <FN> beginning at <Base>...
  ' <Cnt> is just for the information report
  
  Dim Recurse As Long, IFP As Integer, MyByte As Byte
  
  frmInit.lblStatus = "EPROM joiner: reading image, part " & _
                      CStr(Cnt) & "..."
  frmInit.lblStatus.Refresh
  
  IFP = FreeFile
  Open Setup_RebuildFile(FN) For Binary Access Read As IFP
  
  For Recurse = 0 To (FLen - 1)
    Get #IFP, , MyByte
    ROM(Base + Recurse) = MyByte
  Next
  
  Close #IFP
End Sub


Public Sub Memory_EnquireEPROM()
  Dim MyTxt As String
  
  MyTxt = "The EPROM validation code is &" & Right$("0000" & _
          Hex$(ROMCHK), 4) & "." & vbCrLf & vbCrLf

  Select Case ROMCHK
    Case &H31F0&:
      MyTxt = MyTxt & "This EPROM is recognised as the Acorn " & _
              "FileStore E01 firmware version 1.31 (1988)."
                  
    Case &HFE41&:
      MyTxt = MyTxt & "This EPROM is recognised as the Acorn " & _
              "FileStore E01S firmware version 1.33 (1988)."
            
    Case &HE529&:
      MyTxt = MyTxt & "This EPROM is recognised as the Acorn " & _
              "FileStore E01S firmware version 1.40 (1989)."
      
    Case Else:
      MyTxt = MyTxt & "This EPROM is NOT recognised." & vbCrLf & _
              "I would appreciate if you could send me a copy " & _
              "of this firmware so it can be added to the " & _
              "list of versions known..."
  End Select
  
  MsgBox MyTxt, vbInformation + vbOKOnly, _
         "About the current EPROM..."
End Sub


Public Function Memory_LookupAnno(Addr As Long) As String
  Dim Recurse As Integer
  
  For Recurse = 1 To DissAnnoCnt
    If (DissAnno(Recurse).Addr = Addr) Then
      Memory_LookupAnno = DissAnno(Recurse).Anno
      Exit Function
    End If
  Next
End Function


Public Function Memory_ReadEPROMLatch() As Boolean
  ' for disassembly to flag if we're reading ROM or RAM...
  
  Memory_ReadEPROMLatch = MemEPROMLatch
End Function
